
import warnings
import numpy as np
import pandas as pd
from ..GWADataLoader import GWADataLoader


class PhenotypeSimulator(GWADataLoader):
    """
    A wrapper class that supports simulating complex traits with a variety of
    genetic architectures and heritability values, using the standard linear model. The
    basic implementation supports simulating effect sizes from a sparse Gaussian mixture density,
    allowing some variants to have larger effects than others. The class also supports simulating
    binary phenotypes (case-control) by thresholding the continuous phenotype at a specified threshold.

    To be concrete, the generative model for the simulation is as follows:

    1) Simulate the mixture assignment for each variant based on the mixing proportions `pi`.
    2) Simulate the effect sizes for each variant from the corresponding Gaussian density that they were assigned.
    3) Compute the polygenic score for each individual based on the simulated effect sizes.
    4) Simulate the residual component of the phenotype, in such a way that the total heritability is preserved.

    !!! seealso "See Also"
        * [GWADataLoader][magenpy.GWADataLoader.GWADataLoader]

    :ivar pi: The mixing proportions for the Gaussian mixture density.
    :ivar h2: The trait SNP heritability, or proportion of variance explained by SNPs.
    :ivar d: The variance multipliers for each component of the Gaussian mixture density.
    :ivar prevalence: The (disease) prevalence for binary (case-control) phenotypes.
    :ivar per_snp_h2: The per-SNP heritability for each variant in the dataset.
    :ivar per_snp_pi: The per-SNP mixing proportions for each variant in the dataset.
    :ivar beta: The effect sizes for each variant in the dataset.
    :ivar mixture_assignment: The assignment of each variant to a mixture component.

    """

    def __init__(self,
                 bed_files,
                 h2=0.2,
                 pi=0.1,
                 d=(0., 1.),
                 prevalence=0.15,
                 **kwargs):
        """
        Initialize the PhenotypeSimulator object with the necessary parameters.

        :param bed_files: A path (or list of paths) to PLINK BED files containing the genotype information.
        :param h2: The trait SNP heritability, or proportion of variance explained by SNPs.
        :param pi: The mixing proportions for the mixture of Gaussians (our model for the distribution of effect sizes).
        If a float is provided, it is converted to a tuple (1-pi, pi), where pi is the proportion of causal variants.
        :param d:  The variance multipliers for each component of the Gaussian mixture density. By default,
        all components have the same variance multiplier.
        :param prevalence: The (disease) prevalence for binary (case-control) phenotypes.
        """

        super().__init__(bed_files, **kwargs)

        # If pi is float, convert it to a tuple:
        if isinstance(pi, float):
            pi = (1. - pi, pi)

        self.pi = pi
        self.h2 = h2
        self.prevalence = prevalence

        # Sanity checks:
        assert 0. <= self.h2 <= 1.
        assert round(sum(self.pi), 1) == 1.
        assert 0. < self.prevalence < 1.

        self.d = np.array(d)

        self.per_snp_h2 = None
        self.per_snp_pi = None
        self.beta = None
        self.mixture_assignment = None

    @property
    def n_components(self):
        """
        :return: The number of Gaussian mixture components for the effect size distribution.
        """
        return len(self.pi)

    def set_pi(self, new_pi):
        """
        Set the mixture proportions (proportion of variants in each
        Gaussian mixture component).
        """
        self.pi = new_pi
        self.set_per_snp_mixture_probability()

    def set_h2(self, new_h2):
        """
        Set the total heritability (proportion of additive variance due to SNPs) for the trait
        """
        self.h2 = new_h2
        self.set_per_snp_heritability()

    def set_per_snp_mixture_probability(self):
        """
        Set the per-SNP mixing proportions for each variant in the dataset.
        This is a convenience method that may come in handy for more flexible generative models.
        """

        self.per_snp_pi = {}

        for c, c_size in self.shapes.items():
            self.per_snp_pi[c] = np.repeat(np.array([self.pi]), c_size, axis=0)

    def set_per_snp_heritability(self):
        """
        Set the per-SNP heritability (effect size variance) for each variant in the dataset.
        This is a convenience method that may come in handy for more flexible generative models.
        """

        assert self.mixture_assignment is not None

        # Estimate the global sigma_beta_sq based on the
        # pre-specified heritability, the mixture proportions `pi`,
        # and the prior multipliers `d`.

        combined_assignments = np.concatenate([self.mixture_assignment[c] for c in self.chromosomes])

        sigma_beta_sq = self.h2 / (combined_assignments*self.d).sum()

        self.per_snp_h2 = {}

        for c, c_size in self.shapes.items():
            self.per_snp_h2[c] = sigma_beta_sq*self.d[np.where(self.mixture_assignment[c])[1]]

    def get_causal_status(self):
        """
        :return: A dictionary where the keys are the chromosome numbers
        and the values are binary vectors indicating which SNPs are
        causal for the simulated phenotype.

        :raises AssertionError: If the mixture assignment is not set.
        """

        assert self.mixture_assignment is not None

        try:
            zero_index = list(self.d).index(0)
        except ValueError:
            # If all SNPs are causal:
            return {c: np.repeat(True, c_size) for c, c_size in self.shapes.items()}

        causal_status = {}

        for c, mix_a in self.mixture_assignment.items():
            causal_status[c] = np.where(mix_a)[1] != zero_index

        return causal_status

    def set_causal_snps(self, causal_snps):
        """
        A utility method to set the causal variants in the simulation based on an array or
        list of SNPs specified by the user. The method takes an iterable (e.g. list or array) of `causal_snps`
        and then creates a new mixture assignment object where only the `causal_snps`
        contribute to the phenotype.

        :param causal_snps: A list or array of SNP rsIDs.
        :raises ValueError: If all mixture components are causal.

        """

        # Get the index of the mixture component whose multiplier is zero (i.e. the null component):
        try:
            zero_index = list(self.d).index(0)
        except ValueError:
            raise ValueError("Cannot set causal variants when all mixture components are causal. Modify "
                             "the mixture multipliers `d` to enable this functionality.")

        # Get the indices of the non-null mixture components:
        nonzero_indices = [i for i, d in enumerate(self.d) if d != 0.]

        # Get the mixture proportions for the non-null components and normalize them so they sum to 1:
        pis = np.array(self.pi)[nonzero_indices]
        pis /= pis.sum()

        # Initialize new mixture assignment object:
        new_assignment = {c: np.zeros((s, self.n_components)) for c, s in self.shapes.items()}

        from ..utils.compute_utils import intersect_arrays

        n_causal_set = 0

        for c, snps in self.snps.items():

            # Intersect the list of causal SNPs with the SNPs on chromosome `c`:
            snp_idx = intersect_arrays(snps, causal_snps, return_index=True)

            if len(snp_idx) > 0:
                n_causal_set += len(snp_idx)
                # For the causal SNPs, assign them randomly to the causal components
                new_assignment[c][snp_idx, np.random.choice(nonzero_indices,
                                                            size=len(snp_idx),
                                                            p=pis)] = 1
                # For the remaining SNPs, assign them to the null component:
                new_assignment[c][:, zero_index] = np.abs(new_assignment[c].sum(axis=1) - 1)

        if n_causal_set < len(causal_snps):
            warnings.warn("Not all causal SNPs are represented in the genotype matrix! "
                          f"User passed a list of {len(causal_snps)} SNPs, only matched {n_causal_set}.")

        self.set_mixture_assignment(new_assignment)

    def set_mixture_assignment(self, new_assignment):
        """
        Set the mixture assignments according to user-provided dictionary. The mixture
        assignment indicates which mixture component the effect size of a particular
        variant comes from.
        :param new_assignment: A dictionary where the keys are the chromosomes and
        the values are the mixture assignment for each SNP on that chromosome.
        """

        # Check that the shapes match pre-specified information:
        for c, c_size in self.shapes.items():
            assert new_assignment[c].shape == (c_size, self.n_components)

        self.mixture_assignment = new_assignment

    def simulate_mixture_assignment(self):
        """
        Simulate assigning SNPs to the various mixture components
        with probabilities given by mixing proportions `pi`.
        """

        if self.per_snp_pi is None or len(self.per_snp_pi) < 1:
            self.set_per_snp_mixture_probability()

        self.mixture_assignment = {}

        from ..utils.model_utils import multinomial_rvs

        for c, c_size in self.shapes.items():

            self.mixture_assignment[c] = multinomial_rvs(1, self.per_snp_pi[c])

        return self.mixture_assignment

    def set_beta(self, new_beta):
        """
        Set the variant effect sizes (beta) according to user-provided dictionary.

        :param new_beta: A dictionary where the keys are the chromosomes and
        the values are the beta (effect size) for each SNP on that chromosome.
        """

        # Check that the shapes match pre-specified information:
        for c, c_size in self.shapes.items():
            assert len(new_beta[c]) == c_size

        self.beta = new_beta

    def simulate_beta(self):
        """
        Simulate the causal effect size for variants included
        in the dataset. Here, the variant effect size is drawn from
        a Gaussian density with mean zero and scale given by
        the root of per-SNP heritability.
        """

        if self.per_snp_h2 is None or len(self.per_snp_h2) < 1:
            self.set_per_snp_heritability()

        self.beta = {}

        for c, c_size in self.shapes.items():

            self.beta[c] = np.random.normal(loc=0.0,
                                            scale=np.sqrt(self.per_snp_h2[c]),
                                            size=c_size)

        return self.beta

    def simulate_phenotype(self):
        """
        Simulate complex phenotypes for the samples present in the genotype matrix, given their
        genotype information and fixed effect sizes `beta` that were simulated previous steps.

        Given the simulated effect sizes, the phenotype is generated as follows:

        `Y = XB + e`

        Where `Y` is the vector of phenotypes, `X` is the genotype matrix, `B` is the vector of effect sizes,
        and `e` represents the residual effects.
        """

        assert self.beta is not None

        # Compute the polygenic score given the simulated/provided beta:
        pgs = self.score(self.beta)

        # Sample the environmental/residual component:
        e = np.random.normal(loc=0., scale=np.sqrt(1. - self.h2), size=self.sample_size)

        # The final simulated phenotype is a combination of
        # the polygenic score + the residual component:
        y = pgs + e

        if self.phenotype_likelihood == 'binomial':
            # If the simulated phenotype is to be binary,
            # use the threshold model to determine positives/negatives
            # based on the prevalence of the phenotype in the population:
            
            from ..stats.transforms.phenotype import standardize
            
            y = standardize(y)

            from scipy.stats import norm
            
            cutoff = norm.ppf(1. - self.prevalence)
            new_y = np.zeros_like(y, dtype=int)
            new_y[y > cutoff] = 1
        else:
            new_y = y

        self.set_phenotype(new_y)
        
        return new_y

    def simulate(self,
                 reset_beta=True,
                 reset_mixture_assignment=True,
                 perform_gwas=False):
        """
        A convenience method to simulate all the components of the generative model.
        Specifically, the simulation follows the standard linear model, where the phenotype is 
        dependent on the genotype + environmental components that are assumed to be uncorrelated:
        
        `Y = XB + e`
        
        Where `Y` is the vector of phenotypes, `X` is the genotype matrix, `B` is the vector of effect sizes, 
        and `e` represents the residual effects. The generative model proceeds by:
         
         1) Drawing the effect sizes `beta` from a Gaussian mixture density.
         2) Drawing the residual effect from an isotropic Gaussian density.
         3) Setting the phenotype according to the equation above. 
        
        :param reset_beta: If True, reset the effect sizes by drawing new ones from the prior density.
        :param reset_mixture_assignment: If True, reset the assignment of SNPs to mixture components. Set to False
        if you'd like to keep the same configuration of causal SNPs.
        :param perform_gwas: If True, automatically perform genome-wide association on the newly simulated phenotype.
        """

        # Simulate the mixture assignment:
        if self.mixture_assignment is None or reset_mixture_assignment:
            self.simulate_mixture_assignment()

        # Set the per-SNP heritability based on the mixture assignment:
        self.set_per_snp_heritability()

        # Simulate betas based on per-SNP heritability
        if self.beta is None or reset_beta:
            self.simulate_beta()

        # Simulate the phenotype
        self.simulate_phenotype()

        if perform_gwas:
            # Perform genome-wide association testing...
            self.perform_gwas()

    def to_true_beta_table(self, per_chromosome=False):
        """
        Export the simulated true effect sizes and causal status into a pandas dataframe.
        :param per_chromosome: If True, return a dictionary of tables for each chromosome separately.

        :return: A pandas DataFrame with the true effect sizes and causal status for each variant.
        """

        assert self.beta is not None

        eff_tables = {}
        causal_status = self.get_causal_status()

        for c in self.chromosomes:

            eff_tables[c] = pd.DataFrame({
                'CHR': c,
                'SNP': self.genotype[c].snps,
                'A1': self.genotype[c].a1,
                'A2': self.genotype[c].a2,
                'MixtureComponent': np.where(self.mixture_assignment[c] == 1)[1],
                'Heritability': self.per_snp_h2[c],
                'BETA': self.beta[c].flatten(),
                'Causal': causal_status[c],
            })

        if per_chromosome:
            return eff_tables
        else:
            return pd.concat(list(eff_tables.values()))




def standardize(g_mat):
    """
    Standardize the genotype matrix, such that the columns (i.e. variants)
    have zero mean and unit variance.
    :param g_mat: A two-dimensional matrix (numpy, dask, xarray, etc.) where the rows are samples (individuals)
    and the columns are genetic variants.

    :return: The standardized genotype matrix.

    """

    return (g_mat - g_mat.mean(axis=0)) / g_mat.std(axis=0)



from .system_utils import is_cmd_tool, run_shell_script
from magenpy import get_option


class plink2Executor(object):
    """
    A wrapper class for interfacing with the `plink2` command line tool.
    """

    def __init__(self, threads=None, verbose=True):
        """
        Initialize the plink2 executor
        :param threads: The number of threads to use for computations. If None, the number of
        threads will be set to 1.
        :type threads: int or str
        :param verbose: Whether to print the output of the command
        :type verbose: bool
        """

        self.threads = threads or 1

        self.plink2_path = get_option('plink2_path')

        if not is_cmd_tool(self.plink2_path):
            raise Exception(f"Did not find the executable for plink2 at: {self.plink2_path}")

        self.verbose = verbose

    def execute(self, cmd):
        """
        Execute a `plink2` command
        :param cmd: The flags to pass to plink2. For example, ['--bfile', 'file', '--out', 'output']
        :type cmd: list of strings
        :raises CalledProcessError: If the command returns a non-zero exit code
        """

        cmd = [self.plink2_path] + cmd + [f'--threads {self.threads}']

        from subprocess import CalledProcessError

        try:
            run_shell_script(" ".join(cmd))
        except CalledProcessError as e:

            if self.verbose:
                print("Invocation of plink2 returned the following error message:")
                print(e.stderr.decode())

            raise e


class plink1Executor(object):
    """
    A wrapper class for interfacing with the `plink1.9` command line tool.
    """

    def __init__(self, threads=None, verbose=True):
        """
        Initialize the plink1.9 executor
        :param threads: The number of threads to use for computations. If None, the number of
        threads will be set to 1.
        :type threads: int or str
        :param verbose: Whether to print the output of the command
        :type verbose: bool
        """

        self.threads = threads or 1

        self.plink1_path = get_option('plink1.9_path')

        if not is_cmd_tool(self.plink1_path):
            raise Exception(f"Did not find the executable for plink1.9 at: {self.plink1_path}")

        self.verbose = verbose

    def execute(self, cmd):
        """
        Execute a plink command
        :param cmd: The flags to pass to plink. For example, ['--bfile', 'file', '--out', 'output']
        :type cmd: list of strings
        :raises: CalledProcessError if the command fails
        """

        cmd = [self.plink1_path] + cmd + [f'--threads {self.threads}']

        from subprocess import CalledProcessError

        try:
            run_shell_script(" ".join(cmd))
        except CalledProcessError as e:
            if self.verbose:
                print("Invocation of plink returned the following error message:")
                print(e.stderr.decode())
            raise e



import numpy as np


def adjust_for_covariates(phenotype, covariates):
    """
    This function takes a phenotype vector and a matrix of covariates
    and applies covariate correction on the phenotype. Concretely,
    this involves fitting a linear model where the response is the
    phenotype and the predictors are the covariates and then returning
    the residuals.
    :param phenotype: A vector of continuous or quantitative phenotypes.
    :param covariates: A matrix where each row corresponds to an individual
     and each column corresponds to a covariate (e.g. age, sex, PCs, etc.)

     :return: The residuals of the linear model fit.
    """

    import statsmodels.api as sm

    return sm.OLS(phenotype, sm.add_constant(covariates)).fit().resid


def rint(phenotype, offset=3./8):
    """
    Apply Rank-based inverse normal transform on the phenotype.
    :param phenotype: A vector of continuous or quantitative phenotypes.
    :param offset: The offset to use in the INT transformation (Blom's offset by default).

    :return: The RINT-transformed phenotype.
    """

    from scipy.stats import rankdata, norm

    ranked_pheno = rankdata(phenotype, method="average")
    return norm.ppf((ranked_pheno - offset) / (len(ranked_pheno) - 2 * offset + 1))


def detect_outliers(phenotype, sigma_threshold=3, nan_policy='omit'):
    """
    Detect samples with outlier phenotype values.
    This function takes a vector of quantitative phenotypes,
    computes the z-score for every individual, and returns a
    boolean vector indicating whether individual i has phenotype value
    within the specified standard deviations `sigma_threshold`.
    :param phenotype: A numpy vector of continuous or quantitative phenotypes.
    :param sigma_threshold: The multiple of standard deviations or sigmas after
    which we consider the phenotypic value an outlier. Default is 3.
    :param nan_policy: The policy to use when encountering NaN values in the phenotype vector.
    By default, we compute the z-scores ignoring NaN values.

    :return: A boolean array indicating whether the phenotype value is an outlier (i.e.
    True indicates outlier).
    """
    from scipy.stats import zscore
    return np.abs(zscore(phenotype, nan_policy=nan_policy)) > sigma_threshold


def standardize(phenotype):
    """
    Standardize the phenotype vector to have mean zero and unit variance
    :param phenotype: A numpy vector of continuous or quantitative phenotypes.

    :return: The standardized phenotype array.
    """
    return (phenotype - phenotype.mean()) / phenotype.std()


def chained_transform(sample_table,
                      adjust_covariates=False,
                      standardize_phenotype=False,
                      rint_phenotype=False,
                      outlier_sigma_threshold=None,
                      transform_order=('standardize', 'covariate_adjust', 'rint', 'outlier_removal')):
    """
    Apply a chain of transformations to the phenotype vector.
    :param sample_table: An instance of SampleTable that contains phenotype information and other
    covariates about the samples in the dataset.
    :param adjust_covariates: If true, regress out the covariates from the phenotype. By default, we regress out all
    the covariates present in the SampleTable.
    :param standardize_phenotype: If true, standardize the phenotype.
    :param rint_phenotype: If true, apply Rank-based inverse normal transform.
    :param outlier_sigma_threshold: The multiple of standard deviations or sigmas after
    which we consider the phenotypic value an outlier.
    :param transform_order: A tuple specifying the order in which to apply the transformations. By default,
    the order is standardize, covariate_adjust, rint, and outlier_removal.

    :return: The transformed phenotype vector and a boolean mask indicating the samples that were not removed.
    """

    phenotype = sample_table.phenotype
    mask = np.ones_like(phenotype, dtype=bool)

    if sample_table.phenotype_likelihood != 'binomial':
        for transform in transform_order:

            if transform == 'standardize':
                # Standardize the phenotype:
                if standardize_phenotype:
                    phenotype = standardize(phenotype)

            elif transform == 'covariate_adjust':
                # Adjust the phenotype for a set of covariates:
                if adjust_covariates:
                    phenotype = adjust_for_covariates(phenotype, sample_table.get_covariates_matrix()[mask, :])

            elif transform == 'rint':
                # Apply Rank-based inverse normal transform (RINT) to the phenotype:
                if rint_phenotype:
                    phenotype = rint(phenotype)

            elif transform == 'outlier_removal':
                # Remove outlier samples whose phenotypes are more than `threshold` standard deviations from the mean:
                if outlier_sigma_threshold is not None:
                    # Find outliers:
                    mask = ~detect_outliers(phenotype, outlier_sigma_threshold)
                    # Filter phenotype vector to exclude outliers:
                    phenotype = phenotype[mask]

    return phenotype, mask



from ..LDMatrix import LDMatrix
import matplotlib.pylab as plt
import numpy as np


def plot_ld_matrix(ldm: LDMatrix,
                   variant_subset=None,
                   start_row=None,
                   end_row=None,
                   symmetric=False,
                   cmap='RdBu',
                   include_colorbar=True):
    """
    Plot a heatmap representing the LD matrix or portions of it.

    :param ldm: An instance of `LDMatrix`.
    :param variant_subset: A list of variant rsIDs to subset the LD matrix.
    :param start_row: The starting row index for the LD matrix plot.
    :param end_row: The ending row index (not inclusive) for the LD matrix plot.
    :param symmetric: If True, plot the symmetric version of the LD matrix.
    Otherwise, plot the upper triangular part.
    :param cmap: The color map for the LD matrix plot.
    :param include_colorbar: If True, include a colorbar in the plot.
    """

    curr_mask = None

    if variant_subset is not None:
        curr_mask = ldm.get_mask()
        ldm.reset_mask()
        ldm.filter_snps(variant_subset)

    ld_mat = ldm.load_data(start_row=start_row,
                           end_row=end_row,
                           return_symmetric=symmetric,
                           return_square=True,
                           return_as_csr=True,
                           dtype=np.float32).todense()

    plt.imshow(ld_mat, cmap=cmap, vmin=-1., vmax=1., interpolation='none')

    if include_colorbar:
        plt.colorbar()

    # Reset the original mask:
    if curr_mask is not None:
        ldm.set_mask(curr_mask)

    plt.axis('off')



import numpy as np
from ...GWADataLoader import GWADataLoader


def simple_ldsc(gdl: GWADataLoader):
    """
    Provides an estimate of SNP heritability from summary statistics using
    a simplified version of the LD Score Regression framework.
    E[X_j^2] = h^2*l_j + int
    Where the response is the Chi-Squared statistic for SNP j
    and the variable is its LD score.

    :param gdl: An instance of `GWADataLoader` with the LD information and
    summary statistics initialized properly.

    :return: The estimated SNP heritability.
    """

    # Check data types:
    assert gdl.ld is not None and gdl.sumstats_table is not None

    ld_score = []
    chi_sq = []
    sample_size = []

    for c in gdl.chromosomes:
        ld_score.append(gdl.ld[c].ld_score)
        chi_sq.append(gdl.sumstats_table[c].get_chisq_statistic())
        sample_size.append(gdl.sumstats_table[c].n_per_snp.max())

    ld_score = np.concatenate(ld_score)
    chi_sq = np.concatenate(chi_sq)
    sample_size = max(sample_size)

    return (chi_sq.mean() - 1.) * len(ld_score) / (ld_score.mean() * sample_size)


class LDSCRegression(object):
    """
    Perform LD Score Regression using the jackknife method.
    """

    def __init__(self, gdl: GWADataLoader, n_blocks=200, max_chisq=None):
        """
        :param gdl: An instance of GWADataLoader
        :param n_blocks: The number of blocks to use for the jackknife method.
        :param max_chisq: The maximum Chi-Squared statistic to consider.
        """

        self.gdl = gdl
        self.n_blocks = n_blocks

        # ...

    def fit(self):
        """
        Perform LD Score Regression estimation using the jackknife method.

        :raises NotImplementedError: If method is not implemented.
        """

        raise NotImplementedError




import glob
import configparser

# Data structures:
from .AnnotationMatrix import AnnotationMatrix
from .LDMatrix import LDMatrix
from .GWADataLoader import GWADataLoader
from .SumstatsTable import SumstatsTable
from .SampleTable import SampleTable

# Simulation:

from .simulation.PhenotypeSimulator import PhenotypeSimulator

# Data utilities:

from .utils.data_utils import *

__version__ = '0.1.5'
__release_date__ = 'April 2025'


def make_ascii_logo(desc=None, left_padding=None):
    """
    Generate an ASCII logo for the magenpy package.
    :param desc: A string description to be added below the logo.
    :param left_padding: Padding to the left of the logo.

    :return: A string containing the ASCII logo.
    """

    logo = r"""
 _ __ ___   __ _  __ _  ___ _ __  _ __  _   _ 
| '_ ` _ \ / _` |/ _` |/ _ \ '_ \| '_ \| | | |
| | | | | | (_| | (_| |  __/ | | | |_) | |_| |
|_| |_| |_|\__,_|\__, |\___|_| |_| .__/ \__, |
                 |___/           |_|    |___/
    """

    lines = logo.replace(' ', '\u2001').splitlines()[1:]
    lines.append("Modeling and Analysis of Genetics data in python")
    lines.append(f"Version: {__version__} | Release date: {__release_date__}")
    lines.append("Author: Shadi Zabad, McGill University")

    # Find the maximum length of the lines
    max_len = max([len(l) for l in lines])
    if desc is not None:
        max_len = max(max_len, len(desc))

    # Pad the lines to the same length
    for i, l in enumerate(lines):
        lines[i] = l.center(max_len)

    # Add separators at the top and bottom
    lines.insert(0, '*' * max_len)
    lines.append('*' * max_len)

    if desc is not None:
        lines.append(desc.center(max_len))

    if left_padding is not None:
        for i, l in enumerate(lines):
            lines[i] = '\u2001' * left_padding + l

    return "\n".join(lines)


config = configparser.ConfigParser()
config.read(glob.glob(osp.join(osp.dirname(__file__), 'config/*.ini')))


def print_options():
    """
    Print the options stored in the configuration file
    """
    for sec in config.sections() + [config.default_section]:
        print("-> Section:", sec)
        for key in config[sec]:
            print(f"---> {key}: {config[sec][key]}")


def get_option(key):
    """
    Get the option associated with a given key
    """
    try:
        return config['USER'][key]
    except KeyError:
        return config['DEFAULT'][key]


def set_option(key, value):
    """
    Set an option in the configuration file by providing a key and a value
    """
    if 'USER' in config.sections():
        config['USER'][key] = value
    else:
        config['USER'] = {key: value}

    with open(osp.join(osp.dirname(__file__), 'config/paths.ini'), 'w') as configfile:
        config.write(configfile)



from tqdm import tqdm
import numpy as np
import pandas as pd


def match_chromosomes(chrom_1, chrom_2, check_patterns=('chr_', 'chr:', 'chr'), return_both=False):
    """
    Given two lists of chromosome IDs, this function returns the
    chromosomes that are common to both lists. By default, the returned chromosomes
    follow the data type and order of the first list. If `return_both` is set to True,
    the function returns the common chromosomes in both lists.

    The function also accounts for common ways to encode chromosomes, such as
    chr18, chr_18, 18, etc.

    :param chrom_1: A list or numpy array of chromosome IDs
    :param chrom_2: A list or numpy array of chromosome IDs
    :param check_patterns: A list of patterns to check for and replace in the chromosome IDs
    :param return_both: If True, return the common chromosomes in both lists
    """

    chrom_1 = np.array(list(chrom_1))
    chrom_2 = np.array(list(chrom_2))

    # First, convert the data types to strings:
    chr1_str = chrom_1.astype(str)
    chr2_str = chrom_2.astype(str)

    _, chr1_idx, chr2_idx = np.intersect1d(chr1_str, chr2_str, return_indices=True)

    if len(chr1_idx) < 1:
        # Replace patterns
        for pattern in check_patterns:
            chr1_str = np.char.replace(chr1_str, pattern, '')
            chr2_str = np.char.replace(chr2_str, pattern, '')

        _, chr1_idx, chr2_idx = np.intersect1d(chr1_str, chr2_str, return_indices=True)

    if len(chr1_idx) < 1:
        if return_both:
            return [], []
        else:
            return []
    else:
        if return_both:
            return chrom_1[chr1_idx], chrom_2[chr2_idx]
        else:
            return chrom_1[chr1_idx]


def merge_snp_tables(ref_table,
                     alt_table,
                     how='inner',
                     on='auto',
                     signed_statistics=('BETA', 'STD_BETA', 'Z'),
                     drop_duplicates=True,
                     correct_flips=True,
                     return_ref_indices=False,
                     return_alt_indices=False):
    """
    This function takes a reference SNP table with at least 3 columns ('SNP', 'A1', `A2`)
    and matches it with an alternative table that also has these 3 columns defined. In the most recent
    implementation, we allow users to merge on any set of columns that they wish by specifying the `on`
    parameter. For example, instead of `SNP`, the user can join the SNP tables on `CHR` and `POS`, the
    chromosome number and base pair position of the SNP.

    The manner in which the join operation takes place depends on the `how` argument.
    Currently, the function supports `inner` and `left` joins.

    The function removes duplicates if `drop_duplicates` parameter is set to True

    If `correct_flips` is set to True, the function will correct summary statistics in
    the alternative table `alt_table` (e.g. BETA, MAF) based whether the A1 alleles agree between the two tables.

    :param ref_table: The reference table (pandas dataframe).
    :param alt_table: The alternative table (pandas dataframe)
    :param how: `inner` or `left`
    :param on: Which columns to use as anchors when merging. By default, we automatically infer which columns
    to use, but the user can specify this directly. When `on='auto'`, we try to use `SNP` (i.e. rsID) if available.
    If not, we use `['CHR', 'POS']`. If neither are available, we raise a ValueError.
    :param signed_statistics: The columns with signed statistics to flip if `correct_flips` is set to True.
    :param drop_duplicates: Drop duplicate SNPs
    :param correct_flips: Correct SNP summary statistics that depend on status of alternative allele
    :param return_ref_indices: Return the indices of the remaining entries in the reference table before merging.
    :param return_alt_indices: Return the indices of the remaining entries in the alternative table before merging.
    """

    # Sanity checking steps:
    assert how in ('left', 'inner')
    for tab in (ref_table, alt_table):
        assert isinstance(tab, pd.DataFrame)
        if not all([col in tab.columns for col in ('A1', 'A2')]):
            raise ValueError("To merge SNP tables, we require that the columns `A1` and `A2` are present.")

    # Determine which columns to merge on:
    if on == 'auto':
        # Check that the `SNP` column is present in both tables:
        if all(['SNP' in tab.columns for tab in (ref_table, alt_table)]):
            on = ['SNP']
        # Check that the `CHR`, `POS` columns are present in both tables:
        elif all([col in tab.columns for col in ('CHR', 'POS') for tab in (ref_table, alt_table)]):
            on = ['CHR', 'POS']
        else:
            raise ValueError("Cannot merge SNP tables without specifying which columns to merge on.")
    elif isinstance(on, str):
        on = [on]

    ref_include = on + ['A1', 'A2']

    if return_ref_indices:
        ref_table.reset_index(inplace=True, names='REF_IDX')
        ref_include += ['REF_IDX']
    if return_alt_indices:
        alt_table.reset_index(inplace=True, names='ALT_IDX')

    merged_table = ref_table[ref_include].merge(alt_table, how=how, on=on)

    if drop_duplicates:
        merged_table.drop_duplicates(inplace=True, subset=on)

    if how == 'left':
        merged_table['A1_y'] = merged_table['A1_y'].fillna(merged_table['A1_x'])
        merged_table['A2_y'] = merged_table['A2_y'].fillna(merged_table['A2_x'])

    # Assign A1 to be the one derived from the reference table:
    merged_table['A1'] = merged_table['A1_x']
    merged_table['A2'] = merged_table['A2_x']

    # Detect cases where the correct allele is specified in both tables:
    matching_allele = np.all(merged_table[['A1_x', 'A2_x']].values == merged_table[['A1_y', 'A2_y']].values, axis=1)

    # Detect cases where the effect and reference alleles are flipped:
    flip = np.all(merged_table[['A2_x', 'A1_x']].values == merged_table[['A1_y', 'A2_y']].values, axis=1)

    if how == 'inner':
        # Variants to keep:
        if correct_flips:
            keep_snps = matching_allele | flip
        else:
            keep_snps = matching_allele

        # Keep only SNPs with matching alleles or SNPs with flipped alleles:
        merged_table = merged_table.loc[keep_snps, ]
        flip = flip[keep_snps]

    if correct_flips:

        flip = flip.astype(int)
        num_flips = flip.sum()

        if num_flips > 0:

            # If the user provided a single signed statistic as a string, convert to list first:
            if isinstance(signed_statistics, str):
                signed_statistics = [signed_statistics]

            # Loop over the signed statistics and correct them:
            for s_stat in signed_statistics:
                if s_stat in merged_table:
                    merged_table[s_stat] = (-2. * flip + 1.) * merged_table[s_stat]

            # Correct MAF:
            if 'MAF' in merged_table:
                merged_table['MAF'] = np.abs(flip - merged_table['MAF'])

    merged_table.drop(['A1_x', 'A1_y', 'A2_x', 'A2_y'], axis=1, inplace=True)

    return merged_table


def sumstats_train_test_split(gdl, prop_train=0.9, **kwargs):
    """
    Perform a train-test split on the GWAS summary statistics data.
    This function implemented the PUMAS procedure described in

    > Zhao, Z., Yi, Y., Song, J. et al. PUMAS: fine-tuning polygenic risk scores with GWAS summary statistics.
    Genome Biol 22, 257 (2021). https://doi.org/10.1186/s13059-021-02479-9

    Specifically, the function takes harmonized LD and summary statistics data (in the form of a
    `GWADataLoader` object) and samples the marginal beta values for the training set, conditional
    on the LD matrix and the proportion of training data specified by `prop_train`.

    :param gdl: A `GWADataLoader` object containing the harmonized GWAS summary statistics and LD data.
    :param prop_train: The proportion of training data to sample.
    :param kwargs: Additional keyword arguments to pass to the
    `multivariate_normal_conditional_sampling` function.

    :return: A dictionary with the sampled beta values for the training and test sets.
    """

    # Sanity checks:
    assert 0. < prop_train < 1., "The proportion of training data must be between 0 and 1."
    assert gdl.sumstats_table is not None, "The GWADataLoader object must have summary statistics initialized."
    assert gdl.ld is not None, "The GWADataLoader object must have LD matrices initialized."

    from ..stats.ld.utils import multivariate_normal_conditional_sampling

    prop_test = 1. - prop_train

    results = {}

    for chrom in gdl.chromosomes:

        assert gdl.ld[chrom].n_snps == gdl.sumstats_table[chrom].n_snps, (
            "The number of SNPs in the LD matrix and the summary statistics table must match. Invoke the "
            "`harmonize_data` method on the GWADataLoader object to ensure that the data is harmonized."
        )

        n_per_snp = gdl.sumstats_table[chrom].n_per_snp
        n = n_per_snp.max()  # Get the GWAS sample size

        # The covariance scale is computed based on the proportion of training data:
        cov_scale = 1. / (prop_train * n) - (1. / n)
        # Use the standardized marginal beta as the mean:
        mean = gdl.sumstats_table[chrom].standardized_marginal_beta

        # Sample the training beta values:
        sampled_train_beta = multivariate_normal_conditional_sampling(
            gdl.ld[chrom],
            mean=mean,
            cov_scale=cov_scale,
            **kwargs
        )

        # Calculate the test beta values:
        sampled_test_beta = mean * (1. / prop_test) - sampled_train_beta * (prop_train / prop_test)

        # Store the results:
        results[chrom] = {
            'train_beta': sampled_train_beta,
            'train_n': n_per_snp * prop_train,
            'test_beta': sampled_test_beta,
            'test_n': n_per_snp * prop_test
        }

    return results


def map_variants_to_genomic_blocks(variant_table,
                                   block_table,
                                   variant_pos_col='POS',
                                   block_boundary_cols=('block_start', 'block_end'),
                                   filter_unmatched=False):
    """
    Merge a variant table with a genomic block table. This function assumes that the
    variant table includes a column with the positions of the variants `POS` and the block table
    contains the start and end positions of each block.

     !!! warning
         This function assumes that the tables contains data for a single chromosome only.

    :param variant_table: A pandas dataframe with variant information
    :param block_table: A pandas dataframe with block information
    :param variant_pos_col: The name of the column in the variant table that contains the positions. By default,
    this is set to `POS`.
    :param block_boundary_cols: A tuple of two strings specifying the column names in the block table that contain
    the start and end positions of the blocks. By default, this is set to `('block_start', 'block_end')`.
    :param filter_unmatched: If True, filter out variants that were not matched to a block.
    """

    # Sanity checks:
    assert variant_pos_col in variant_table.columns
    assert all([col in block_table.columns for col in block_boundary_cols])

    # Sort the variant table by position:
    variant_table.sort_values(variant_pos_col, inplace=True)

    # Merge the two dataframes to assign each SNP to its corresponding block:
    merged_df = pd.merge_asof(variant_table, block_table,
                              left_on=variant_pos_col, right_on=block_boundary_cols[0],
                              direction='backward')

    if filter_unmatched:
        # Filter merged_df to only include variants that were matched properly with a block:
        merged_df = merged_df.loc[(merged_df[variant_pos_col] >= merged_df[block_boundary_cols[0]]) &
                                  (merged_df[variant_pos_col] < merged_df[block_boundary_cols[1]])]

    return merged_df


def identify_mismatched_snps(gdl,
                             chrom=None,
                             n_iter=10,
                             G=100,
                             p_dentist_threshold=5e-8,
                             p_gwas_threshold=1e-2,
                             rsq_threshold=.95,
                             max_removed_per_iter=.005):
    """
    This function implements a simple quality control procedures
    that checks that the GWAS summary statistics (Z-scores)
    are consistent with the LD reference panel. This is done
    using a simplified version of the framework outlined in the DENTIST paper:

    Improved analyses of GWAS summary statistics by reducing data heterogeneity and errors
    Chen et al. 2021

    Compared to DENTIST, the simplifications we make are:
        -   For each SNP, we sample one neighboring SNP at a time and compute the T statistic
            using that neighbor's information. The benefit of this is that we don't need to
            invert any matrices, so it's a fast operation to run.
        -   To arrive at a more robust estimate, we sample up to `k` neighbors and average
            the T-statistic across those `k` neighbors.

    NOTE: May need to re-implement this to apply some of the constraints genome-wide
    rather than on a per-chromosome basis.

    :param gdl: A `GWADataLoader` object
    :param chrom: Perform checking only on chromosome `chrom`
    :param n_iter: Number of iterations
    :param G: The number of neighboring SNPs to sample (default: 100)
    :param p_dentist_threshold: The Bonferroni-corrected P-value threshold (default: 5e-8)
    :param p_gwas_threshold: The nominal GWAS P-value threshold for partitioning variants (default: 1e-2)
    :param rsq_threshold: The R^2 threshold to select neighbors (neighbor's squared
    correlation coefficient must be less than specified threshold).
    :param max_removed_per_iter: The maximum proportion of variants removed in each iteration
    """

    # Import required modules / functions:
    from scipy import stats

    # Data preparation:
    if chrom is None:
        chromosomes = gdl.chromosomes
    else:
        chromosomes = [chrom]

    shapes = gdl.shapes
    mismatched_dict = {c: np.repeat(False, gdl.shapes[c])
                       for c in chromosomes}

    p_gwas_above_thres = {c: gdl.sumstats_table[c].p_value > p_gwas_threshold for c in chromosomes}
    gwas_thres_size = {c: p.sum() for c, p in p_gwas_above_thres.items()}
    converged = {c: False for c in chromosomes}

    for j in tqdm(range(n_iter),
                  total=n_iter,
                  desc="Identifying mismatched SNPs..."):

        for chrom in chromosomes:

            if converged[chrom]:
                continue

            ld_bounds = gdl.ld[chrom].get_masked_boundaries()
            z = gdl.z_scores[chrom]  # Obtain the z-scores
            t = np.zeros_like(z)

            # Loop over the LD matrix:
            for i, r in enumerate(gdl.ld[chrom]):

                # If the number of neighbors is less than 10, skip...
                if mismatched_dict[chrom][i] or len(r) < 10:
                    continue

                start_idx = ld_bounds[0, i]
                # Select neighbors randomly
                # Note: We are excluding neighbors whose squared correlation coefficient
                # is greater than pre-specified threshold:
                p = (np.array(r)**2 < rsq_threshold).astype(float)
                p /= p.sum()

                neighbor_idx = np.random.choice(len(r), p=p, size=G)
                neighbor_r = np.array(r)[neighbor_idx]

                # Predict the z-score of snp i, given the z-scores of its neighbors:
                pred_z = neighbor_r*z[start_idx + neighbor_idx]

                # Compute the Td statistic for each neighbor and average:
                t[i] = ((z[i] - pred_z) ** 2 / (1. - neighbor_r**2)).mean()

            # Compute the DENTIST p-value assuming a Chi-Square distribution with 1 dof.
            dentist_pval = 1. - stats.chi2.cdf(t, 1)
            # Use a Bonferroni correction to select mismatched SNPs:
            mismatched_snps = dentist_pval < p_dentist_threshold

            if mismatched_snps.sum() < 1:
                # If no new mismatched SNPs are identified, stop iterating...
                converged[chrom] = True
            elif j == n_iter - 1:
                # If this is the last iteration, take all identified SNPs
                mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatched_snps)
            else:

                # Otherwise, we will perform the iterative filtering procedure
                # by splitting variants based on their GWAS p-values:

                # (1) Group S1: SNPs to remove from P_GWAS > threshold:
                mismatch_above_thres = mismatched_snps & p_gwas_above_thres[chrom]
                n_mismatch_above_thres = mismatch_above_thres.sum()
                prop_mismatch_above_thres = n_mismatch_above_thres / gwas_thres_size[chrom]

                if n_mismatch_above_thres < 1:
                    # If no mismatches are detected above the threshold, filter
                    # the mismatches below the threshold and continue...
                    mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatched_snps)
                    continue

                # Sort the DENTIST p-values by index:
                sort_d_pval_idx = np.argsort(dentist_pval)

                if prop_mismatch_above_thres > max_removed_per_iter:
                    idx_to_keep = sort_d_pval_idx[mismatch_above_thres][
                                  int(gwas_thres_size[chrom]*max_removed_per_iter):]
                    mismatch_above_thres[idx_to_keep] = False

                # (2) Group S2: SNPs to remove from P_GWAS < threshold

                # Find mismatched variants below the threshold:
                mismatch_below_thres = mismatched_snps & (~p_gwas_above_thres[chrom])
                n_mismatch_below_thres = mismatch_below_thres.sum()
                prop_mismatch_below_thres = n_mismatch_below_thres / (shapes[chrom] - gwas_thres_size[chrom])

                # For the mismatched variants below the threshold,
                # we remove the same proportion as the variants above the threshold:
                prop_keep_below_thres = min(max_removed_per_iter, prop_mismatch_above_thres)

                if prop_mismatch_below_thres > prop_keep_below_thres:
                    idx_to_keep = sort_d_pval_idx[mismatch_below_thres][
                                  int((shapes[chrom] - gwas_thres_size[chrom]) * prop_keep_below_thres):
                                  ]
                    mismatch_below_thres[idx_to_keep] = False

                # Update the number of variants above the threshold:
                gwas_thres_size[chrom] -= mismatch_above_thres.sum()

                # Update the mismatched dictionary:
                mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatch_below_thres | mismatch_above_thres)

    return mismatched_dict


def get_shared_distance_matrix(tree, tips=None):
    """
    This function takes a Biopython tree and returns the
    shared distance matrix, i.e. for a pair of clades or populations,
    time to most recent common ancestor of the pair minus the time of
    the most recent common ancestor (MRCA).
    """

    tips = tree.get_terminals() if tips is None else tips
    n_tips = len(tips)  # Number of terminal species
    sdist_matrix = np.zeros((n_tips, n_tips))  # Shared distance matrix

    for i in range(n_tips):
        for j in range(i, n_tips):
            if i == j:
                sdist_matrix[i, j] = tree.distance(tree.root, tips[i])
            else:
                mrca = tree.common_ancestor(tips[i], tips[j])
                sdist_matrix[i, j] = sdist_matrix[j, i] = tree.distance(tree.root, mrca)

    return sdist_matrix


def tree_to_rho(tree, min_corr):
    """
    This function takes a Biopython tree and a minimum correlation
    parameter and returns the correlation matrix for the effect sizes
    across populations.

    :param tree: a Biopython Phylo object
    :param min_corr: minimum correlation
    :return:
    """

    max_depth = max(tree.depths().values())
    tree.root.branch_length = min_corr*max_depth / (1. - min_corr)
    max_depth = max(tree.depths().values())

    for c in tree.find_clades():
        c.branch_length /= max_depth

    return tree.root.branch_length + get_shared_distance_matrix(tree)


def quantize(floats, int_dtype=np.int8):
    """
    Quantize floating point numbers to the specified integer type.
    NOTE: Assumes that the floats are in the range [-1, 1].
    :param floats: A numpy array of floats
    :param int_dtype: The integer type to quantize to.
    """

    # Infer the boundaries from the integer type
    info = np.iinfo(int_dtype)

    # NOTE: We add 1 to the info.min here to force the zero point to be exactly at 0.
    # See discussions on Scale Quantization Mapping.

    # Use as much in-place operations as possible
    # (Currently, we copy the data twice)
    scaled_floats = floats*info.max
    np.round(scaled_floats, out=scaled_floats)
    np.clip(scaled_floats, info.min + 1, info.max, out=scaled_floats)

    return scaled_floats.astype(int_dtype)


def dequantize(ints, float_dtype=np.float32):
    """
    Dequantize integers to the specified floating point type.
    NOTE: Assumes original floats are in the range [-1, 1].
    :param ints: A numpy array of integers
    :param float_dtype: The floating point data type to dequantize the integers to.
    """

    # Infer the boundaries from the integer type
    info = np.iinfo(ints.dtype)

    dq = ints.astype(float_dtype)
    dq /= info.max  # in-place multiplication

    return dq


def multinomial_rvs(n, p):
    """
    Copied from Warren Weckesser:
    https://stackoverflow.com/a/55830796

    Sample from the multinomial distribution with multiple p vectors.

    * n must be a scalar.
    * p must an n-dimensional numpy array, n >= 1.  The last axis of p
      holds the sequence of probabilities for a multinomial distribution.

    The return value has the same shape as p.
    """
    count = np.full(p.shape[:-1], n)
    out = np.zeros(p.shape, dtype=int)
    ps = p.cumsum(axis=-1)
    # Conditional probabilities
    with np.errstate(divide='ignore', invalid='ignore'):
        condp = p / ps
    condp[np.isnan(condp)] = 0.0
    for i in range(p.shape[-1]-1, 0, -1):
        binsample = np.random.binomial(count, condp[..., i])
        out[..., i] = binsample
        count -= binsample
    out[..., 0] = count
    return out



from typing import Union
from ..GWADataLoader import GWADataLoader
from ..SumstatsTable import SumstatsTable
import matplotlib.pylab as plt
import pandas as pd
import numpy as np


def manhattan(input_data: Union[GWADataLoader, SumstatsTable, pd.DataFrame],
              x_col=None,
              y_col='NLOG10_PVAL',
              scatter_kwargs=None,
              highlight_snps=None,
              highlight_snps_kwargs=None,
              chrom_sep_color='#f0f0f0',
              add_bonf_line=True,
              bonf_line_kwargs=None):

    """
    Generate Manhattan plot where the x-axis is the genomic position (in BP)
    and the y-axis is the -log10(p-value) or some other statistic of the user's choice.

    :param input_data: An instance of `SumstatsTable` / `GWADataLoader` or a `pandas.DataFrame` from
    which data about the positions of the SNPs and their statistics will be extracted. If `x_col` and `y_col` are
    not provided, the function will assume that the input data contains columns named 'POS' and 'NLOG10_PVAL',
    for the position in base pairs and the -log10(p-value) respectively.
    :param x_col: The column name in the input data that contains the x-axis values. Defaults to None. If the
    user passes `GWADataLoader` or `SumstatsTable` objects, we attempt to extract the x-axis
    values from the 'POS' (position in base pairs) column.
    :param y_col: The column name in the input data that contains the y-axis values (default: 'NLOG10_PVAL').
    :param scatter_kwargs: A dictionary of keyword arguments to pass to the `plt.scatter` function.
    This can be used to customize the appearance of the points on the scatter plot.
    :param highlight_snps: A list or array of SNP rsIDs to highlight on the scatter plot.
    :param highlight_snps_kwargs: A dictionary of keyword arguments to pass to the `plt.scatter` function
    to customize the appearance of the highlighted SNPs.
    :param chrom_sep_color: The color for the chromosome separator block.
    :param add_bonf_line: If True, add a line indicating the Bonferroni significance threshold.
    :param bonf_line_kwargs: The color of the Bonferroni significance threshold line.

    """

    # -------------------------------------------------------
    # Process the input data:

    if isinstance(input_data, SumstatsTable):
        if x_col is None:
            x_col = 'POS'
        input_data = {c: ss.to_table(col_subset=('CHR', 'SNP', x_col, y_col))
                      for c, ss in input_data.split_by_chromosome().items()}
    elif isinstance(input_data, GWADataLoader):
        if x_col is None:
            x_col = 'POS'
        input_data = input_data.to_summary_statistics_table(col_subset=('CHR', 'SNP', x_col, y_col),
                                                            per_chromosome=True)
    elif isinstance(input_data, pd.DataFrame):

        # Sanity checks:
        assert 'CHR' in input_data.columns, "The input data must contain a column named 'CHR'."

        # If the x-axis column is not provided, we assume that the user wishes to plot the
        # variant rank on the x-axis.
        if x_col is not None:
            assert x_col in input_data.columns, f"The input data must contain a column named '{x_col}'."
        else:
            x_col = 'Variant order'
            input_data['Variant order'] = np.arange(1, input_data.shape[0] + 1)

        assert y_col in input_data.columns, f"The input data must contain a column named '{y_col}'."

        if highlight_snps is not None:
            assert 'SNP' in input_data.columns

        input_data = {c: ss for c, ss in input_data.groupby('CHR')}
    else:
        raise ValueError("The input data must be an instance of `SumstatsTable`, `GWADataLoader` "
                         "or a `pandas.DataFrame`.")

    # -------------------------------------------------------
    # Process the SNPs to be highlighted:

    if highlight_snps is not None:
        if isinstance(highlight_snps, list):
            highlight_snps = np.array(highlight_snps)

        highlight_snps = {c: np.in1d(p['SNP'], highlight_snps) for c, p in input_data.items()}

    # -------------------------------------------------------
    # Add custom scatter plot arguments (if not provided)
    if scatter_kwargs is None:
        scatter_kwargs = {'marker': '.', 'alpha': 0.3, 'color': '#808080'}
    else:
        # Only update the keys that are not already present in the dictionary:
        scatter_kwargs = {**scatter_kwargs, **{'marker': '.', 'alpha': 0.3, 'color': '#808080'}}

    if highlight_snps is not None:
        if highlight_snps_kwargs is None:
            highlight_snps_kwargs = {'marker': 'o', 'color': 'red', 'zorder': 2}
        else:
            # Only update the keys that are not already present in the dictionary:
            highlight_snps_kwargs = {**highlight_snps_kwargs, **{'marker': 'o', 'color': 'red', 'zorder': 2}}

    # Add custom Bonferroni line arguments (if not provided)
    if bonf_line_kwargs is None:
        bonf_line_kwargs = {'color': '#b06a7a', 'ls': '--', 'zorder': 1}
    else:
        # Only update the keys that are not already present in the dictionary:
        bonf_line_kwargs = {**bonf_line_kwargs, **{'color': '#b06a7a', 'ls': '--', 'zorder': 1}}

    # -------------------------------------------------------

    if y_col == 'NLOG10_PVAL' and add_bonf_line:
        # Add Bonferroni significance threshold line:
        plt.axhline(-np.log10(0.05 / 1e6), **bonf_line_kwargs)

    # -------------------------------------------------------
    # Iterate through chromosomes and generate scatter plots:

    starting_pos = 0
    ticks = []
    chrom_spacing = .1*min([p[x_col].max() - p[x_col].min() for c, p in input_data.items()])

    unique_chr = sorted(list(input_data.keys()))

    for i, c in enumerate(unique_chr):

        min_pos = input_data[c][x_col].min()
        max_pos = input_data[c][x_col].max()

        xmin = min_pos + starting_pos
        xmax = max_pos + starting_pos

        if i % 2 == 1:
            plt.axvspan(xmin=xmin, xmax=xmax, zorder=0, color=chrom_sep_color)

        ticks.append((xmin + xmax) / 2)

        plt.scatter(input_data[c][x_col] + starting_pos,
                    input_data[c][y_col],
                    **scatter_kwargs)

        if highlight_snps is not None:
            plt.scatter((input_data[c][x_col] + starting_pos)[highlight_snps[c]],
                        input_data[c][y_col][highlight_snps[c]],
                        **highlight_snps_kwargs)

        starting_pos += max_pos + chrom_spacing

    plt.xticks(ticks, unique_chr)

    if x_col == 'POS':
        x_col = 'Genomic Position (BP)'

    plt.xlabel(x_col)

    if y_col == 'NLOG10_PVAL':
        y_col = "$-log_{10}$(p-value)"
    elif y_col == 'PVAL':
        y_col = "p-value"

    plt.ylabel(y_col)
    plt.tight_layout()


def qq_plot(input_data: Union[GWADataLoader, SumstatsTable],
            statistic='p_value'):
    """
    Generate a quantile-quantile (QQ) plot for the GWAS summary statistics.
    The function supports plotting QQ plots for the -log10(p-values) as well as
    the z-score (if available).

    :param input_data: An instance of `SumstatsTable` or `GWADataLoader` from which data about the
    positions of the SNPs will be extracted.
    :param statistic: The statistic to generate the QQ plot for. We currently support `p_value` and `z_score`.
    """

    import scipy.stats as stats

    if statistic == 'p_value':

        if isinstance(input_data, SumstatsTable):
            p_val = input_data.negative_log10_p_value
            m = input_data.m
        elif isinstance(input_data, GWADataLoader):
            p_val = np.concatenate([ss.negative_log10_p_value
                                    for ss in input_data.sumstats_table.values()])
            m = input_data.m
        else:
            raise ValueError("The input data must be an instance of `SumstatsTable` or `GWADataLoader`.")

        plt.scatter(-np.log10(np.arange(1, m + 1) / m), np.sort(p_val)[::-1])

        line = np.linspace(0., p_val.max() + 0.1*p_val.max())
        plt.plot(line, line, c='red')
        plt.xlabel("Expected $-log_{10}$(p-value)")
        plt.ylabel("Observed $-log_{10}$(p-value)")

    elif statistic == 'z_score':
        if isinstance(input_data, SumstatsTable):
            z_scs = input_data.z_score
        elif isinstance(input_data, GWADataLoader):
            z_scs = np.concatenate([ss.z_score for ss in input_data.sumstats_table.values()])
        else:
            raise ValueError("The input data must be an instance of `SumstatsTable` or `GWADataLoader`.")

        stats.probplot(z_scs, dist="norm", plot=plt)
        plt.show()
    else:
        raise NotImplementedError(f"No QQ plot can be generated for the statistic: {statistic}")




import pandas as pd


class AnnotationMatrixParser(object):
    """
    A generic annotation matrix parser class.
    """

    def __init__(self, col_name_converter=None, **read_csv_kwargs):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        SNP features in the annotation matrix.
        :param read_csv_kwargs: Keyword arguments to pass to pandas' `read_csv`.
        """

        self.col_name_converter = col_name_converter
        self.read_csv_kwargs = read_csv_kwargs

        # If the delimiter is not specified, assume whitespace by default:
        if 'sep' not in self.read_csv_kwargs and 'delimiter' not in self.read_csv_kwargs:
            self.read_csv_kwargs['sep'] = r'\s+'

    def parse(self, annotation_file, drop_na=True):
        """
        Parse the annotation matrix file
        :param annotation_file: The path to the annotation file.
        :param drop_na: Drop any entries with missing values.
        """

        try:
            df = pd.read_csv(annotation_file, **self.read_csv_kwargs)
        except Exception as e:
            raise e

        if drop_na:
            df = df.dropna()

        if self.col_name_converter is not None:
            df.rename(columns=self.col_name_converter, inplace=True)

        df.sort_values(['CHR', 'POS'], inplace=True)

        annotations = [ann for ann in df.columns if ann not in ('CHR', 'SNP', 'POS')]

        return df, annotations


class LDSCAnnotationMatrixParser(AnnotationMatrixParser):

    def __init__(self, col_name_converter=None, **read_csv_kwargs):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        SNP features in the annotation matrix.
        :param read_csv_kwargs: Keyword arguments to pass to pandas' read_csv
        """

        super().__init__(col_name_converter, **read_csv_kwargs)
        self.col_name_converter = self.col_name_converter or {}
        self.col_name_converter.update(
            {
                'BP': 'POS'
            }
        )

    def parse(self, annotation_file, drop_na=True):
        """
        Parse the annotation matrix file
        :param annotation_file: The path to the annotation file.
        :param drop_na: Drop any entries with missing values.
        """

        df, annotations = super().parse(annotation_file, drop_na=drop_na)

        df = df.drop(['CM', 'base'], axis=1)
        annotations = [ann for ann in annotations if ann not in ('CM', 'base')]

        return df, annotations


def parse_annotation_bed_file(annot_bed_file):
    """
    Parse an annotation bed file in the format specified by Ensemble:
    https://uswest.ensembl.org/info/website/upload/bed.html

    The file contains 3-12 columns, starting with Chromosome, start_coordinate, end_coordinate, etc.
    After reading the raw file, we let pandas infer whether the file has a header or not and we
    standardize the names of the first 3 columns and convert the chromosome column into an integer.

    :param annot_bed_file: The path to the annotation BED file.
    :type annot_bed_file: str
    """

    try:
        annot_bed = pd.read_csv(annot_bed_file, usecols=[0, 1, 2],
                                sep=r'\s+',
                                names=['CHR', 'Start', 'End'])
    except Exception as e:
        raise e

    annot_bed['CHR'] = annot_bed['CHR'].str.replace('chr', '').astype(int)

    return annot_bed



import numpy as np
import pandas as pd
import os.path as osp
import tempfile
from ...LDMatrix import LDMatrix
from ...utils.system_utils import makedir


class SampleLD(object):
    """
    A basic wrapper class to facilitate computing Linkage-Disequilibrium (LD) matrices.

    Linkage-Disequilibrium (LD) is a measure of the SNP-by-SNP pairwise correlation between
    genetic variants in a population. LD tends to decay with genomic distance, and the rate
    of decay is influenced by many factors. Therefore, LD matrices are often diagonally-dominant.

    This class `SampleLD` provides a basic interface to compute sample correlation coefficient between
     all variants defined in a genotype matrix. The resulting LD matrix is a square and dense matrix.

     For sparse LD matrices, consider using the `WindowedLD`, `ShrinkageLD` or `BlockLD` estimators instead.

     !!! seealso "See Also"
        * [WindowedLD][magenpy.stats.ld.estimator.WindowedLD]
        * [ShrinkageLD][magenpy.stats.ld.estimator.ShrinkageLD]
        * [BlockLD][magenpy.stats.ld.estimator.BlockLD]

     :ivar genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix` or its children.
     :ivar ld_boundaries: The LD boundaries for each variant in the LD matrix.
     :ivar temp_dir: A temporary directory to store intermediate files and results.
     :ivar temp_dir_prefix: A prefix for the temporary directory.

    """

    estimator_id = 'sample'

    def __init__(self, genotype_matrix):
        """
        Initialize the LD estimator with a genotype matrix.
        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        """

        self.genotype_matrix = genotype_matrix
        self.ld_boundaries = None

        self.temp_dir = self.genotype_matrix.temp_dir
        self.temp_dir_prefix = self.genotype_matrix.temp_dir_prefix + 'ld_'

        # Ensure that the genotype matrix has data for a single chromosome only:
        if self.genotype_matrix.chromosome is None:
            raise Exception("`magenpy` does not support computing inter-chromosomal LD matrices! "
                            "You may need to split the genotype matrix by chromosome. "
                            "See GenotypeMatrix.split_by_chromosome().")

    def compute_ld_boundaries(self):
        """
         Compute the Linkage-Disequilibrium (LD) boundaries. LD boundaries define the window
         for which we compute the correlation coefficient between the focal SNP and all other SNPs in
         the genome. Typically, this window is local, since the LD decays exponentially with
         genomic distance.

         The LD boundaries are a 2xM matrix, where M is the number of SNPs on the chromosome.
         The first row contains the start position for the window and the second row contains
         the end position.

         For the sample LD matrix, we simply take the entire square matrix as our window,
         so the start position is 0 and end position is M for all SNPs.

         :return: A 2xM matrix of LD boundaries.
        """

        if self.ld_boundaries is None:

            m = self.genotype_matrix.n_snps
            self.ld_boundaries = np.array((np.zeros(m),
                                           np.ones(m)*m)).astype(np.int64)

        return self.ld_boundaries

    def compute_plink_window_thresholds(self, ld_boundaries=None):
        """
        Computes the LD window thresholds to pass to plink1.9 for computing LD matrices.
        Unfortunately, plink1.9 sets some default values for the window size and it
        is important to set all the thresholds to obtain the desired shape for the
        LD matrix.

        :param ld_boundaries: The LD boundaries for which to compute the thresholds. If not passed,
        we compute the LD boundaries using the `compute_ld_boundaries` method.

        :return: A dictionary containing the window size thresholds for plink1.9.

        """

        if ld_boundaries is None:
            ld_boundaries = self.compute_ld_boundaries()

        threshold_dict = {}

        # (1) Determine maximum window size (Maximum number of neighbors on each side):
        try:
            threshold_dict['window_size'] = getattr(self, "window_size")
            assert threshold_dict['window_size'] is not None
        except (AttributeError, AssertionError):
            threshold_dict['window_size'] = np.abs(ld_boundaries -
                                                   np.arange(ld_boundaries.shape[1])).max()

        # (2) Determine the maximum window size in kilobases + Centi Morgan (if available):

        positional_bounds = np.array([ld_boundaries[0, :], ld_boundaries[1, :] - 1])

        try:
            threshold_dict['kb_window_size'] = getattr(self, "kb_window_size")
            assert threshold_dict['kb_window_size'] is not None
        except (AttributeError, AssertionError):
            kb_pos = .001 * self.genotype_matrix.bp_pos
            kb_bounds = kb_pos[positional_bounds]
            threshold_dict['kb_window_size'] = np.abs(kb_bounds - kb_pos).max()

        # (3) centi Morgan:
        try:
            threshold_dict['cm_window_size'] = getattr(self, "cm_window_size")
            assert threshold_dict['cm_window_size'] is not None
        except (AttributeError, AssertionError):
            try:
                # Checks if cm_pos is available in the genotype matrix:
                cm_pos = self.genotype_matrix.cm_pos
                cm_bounds = self.genotype_matrix.cm_pos[positional_bounds]
                threshold_dict['cm_window_size'] = np.abs(cm_bounds - cm_pos).max()
            except KeyError:
                del threshold_dict['cm_window_size']

        if self.estimator_id == 'sample':
            # If we're using the sample estimator, then expand the plink thresholds slightly
            # to avoid the program dropping LD between variants at the extreme ends:
            for key, val in threshold_dict.items():
                threshold_dict[key] = val*1.05
                if isinstance(val, int):
                    threshold_dict[key] = int(threshold_dict[key])

        return threshold_dict

    def compute(self,
                output_dir,
                overwrite=True,
                delete_original=True,
                dtype='int16',
                compressor_name='zstd',
                compression_level=7,
                compute_spectral_properties=False) -> LDMatrix:
        """
        A utility method to compute the LD matrix and store in Zarr array format.
        The computes the LD matrix and stores it in Zarr array format, set its attributes,
        and performs simple validation at the end.

        :param output_dir: The path where to store the resulting LD matrix.
        :param overwrite: If True, overwrite any existing LD matrices in `temp_dir` and `output_dir`.
        :param delete_original: If True, deletes dense or intermediate LD matrices generated along the way.
        :param dtype: The data type for the entries of the LD matrix (supported data types are float32, float64
        and integer quantized data types int8 and int16).
        :param compressor_name: The name of the compressor to use for the LD matrix.
        :param compression_level: The compression level to use for the LD matrix (1-9).
        :param compute_spectral_properties: If True, compute and store information about the eigenvalues of
        the LD matrix.

        :return: An instance of `LDMatrix` containing the computed LD matrix.

        """

        from .utils import compute_ld_xarray, compute_ld_plink1p9
        from ...GenotypeMatrix import xarrayGenotypeMatrix, plinkBEDGenotypeMatrix

        assert str(dtype) in ('float32', 'float64', 'int8', 'int16')

        # Create the temporary directory using tempfile:
        temp_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix=self.temp_dir_prefix)

        ld_boundaries = self.compute_ld_boundaries()

        if isinstance(self.genotype_matrix, xarrayGenotypeMatrix):
            ld_mat = compute_ld_xarray(self.genotype_matrix,
                                       ld_boundaries,
                                       output_dir,
                                       temp_dir=temp_dir.name,
                                       overwrite=overwrite,
                                       delete_original=delete_original,
                                       dtype=dtype,
                                       compressor_name=compressor_name,
                                       compression_level=compression_level)
        elif isinstance(self.genotype_matrix, plinkBEDGenotypeMatrix):

            # Compute the window size thresholds to pass to plink 1.9:
            window_size_thersh = self.compute_plink_window_thresholds(ld_boundaries)

            ld_mat = compute_ld_plink1p9(self.genotype_matrix,
                                         ld_boundaries,
                                         output_dir,
                                         window_size_thersh,
                                         trim_boundaries=self.estimator_id not in ('sample', 'windowed'),
                                         temp_dir=temp_dir.name,
                                         overwrite=overwrite,
                                         dtype=dtype,
                                         compressor_name=compressor_name,
                                         compression_level=compression_level)
        else:
            raise NotImplementedError

        # Add attributes to the LDMatrix object:
        ld_mat.set_store_attr('Chromosome', int(self.genotype_matrix.chromosome))
        ld_mat.set_store_attr('Sample size', int(self.genotype_matrix.sample_size))
        ld_mat.set_store_attr('LD estimator', 'sample')

        if self.genotype_matrix.genome_build is not None:
            ld_mat.set_store_attr('Genome build', self.genotype_matrix.genome_build)

        ld_mat.set_metadata('snps', self.genotype_matrix.snps, overwrite=overwrite)
        ld_mat.set_metadata('bp', self.genotype_matrix.bp_pos, overwrite=overwrite)
        ld_mat.set_metadata('maf', self.genotype_matrix.maf, overwrite=overwrite)
        ld_mat.set_metadata('a1', self.genotype_matrix.a1, overwrite=overwrite)
        ld_mat.set_metadata('a2', self.genotype_matrix.a2, overwrite=overwrite)

        try:
            ld_mat.set_metadata('cm', self.genotype_matrix.cm_pos, overwrite=overwrite)
        except KeyError:
            pass

        ld_mat.set_metadata('ldscore', ld_mat.compute_ld_scores(), overwrite=overwrite)

        if compute_spectral_properties:

            extreme_eigs = ld_mat.estimate_extremal_eigenvalues()

            ld_mat.set_store_attr('Spectral properties', {
                'Extremal': extreme_eigs
            })

        if ld_mat.validate_ld_matrix():
            temp_dir.cleanup()
            return ld_mat


class WindowedLD(SampleLD):
    """
    A wrapper class to facilitate computing windowed Linkage-Disequilibrium (LD) matrices.
    Windowed LD matrices only record pairwise correlations between variants that are within a certain
    distance of each other along the chromosome. This is useful for reducing the memory requirements
    and noise in the LD matrix.

    The `WindowedLD` estimator supports a variety of ways for defining the window size:

    * `window_size`: The number of neighboring SNPs to consider on each side when computing LD.
    * `kb_window_size`: The maximum distance in kilobases to consider when computing LD.
    * `cm_window_size`: The maximum distance in centi Morgan to consider when computing LD.

    The LD boundaries computed here are the intersection of the windows defined by the window size around
    each SNP (`window_size`), the window size in kilobases (`kb_window_size`), and the window size in centi Morgan
    (`cm_window_size`).

    !!! seealso "See Also"
        * [WindowedLD][magenpy.stats.ld.estimator.ShrinkageLD]
        * [BlockLD][magenpy.stats.ld.estimator.BlockLD]

    :ivar genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
    :ivar ld_boundaries: The LD boundaries for each variant in the LD matrix.
    :ivar window_size: The number of neighboring SNPs to consider on each side when computing LD.
    :ivar kb_window_size: The maximum distance in kilobases to consider when computing LD.
    :ivar cm_window_size: The maximum distance in centi Morgan to consider when computing LD.

    """

    estimator_id = 'windowed'

    def __init__(self,
                 genotype_matrix,
                 window_size=None,
                 kb_window_size=None,
                 cm_window_size=None):
        """

        Initialize the windowed LD estimator with a genotype matrix and window size parameters.

        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param window_size: The number of neighboring SNPs to consider on each side when computing LD.
        :param kb_window_size: The maximum distance in kilobases to consider when computing LD.
        :param cm_window_size: The maximum distance in centi Morgan to consider when computing LD.
        """

        super().__init__(genotype_matrix=genotype_matrix)

        assert not all([w is None for w in (window_size, kb_window_size, cm_window_size)])

        self.window_size = window_size
        self.kb_window_size = kb_window_size
        self.cm_window_size = cm_window_size

    def compute_ld_boundaries(self):
        """
         Compute the windowed Linkage-Disequilibrium (LD) boundaries.
         The LD boundaries computed here are the intersection of the windows defined by
         the window size around each SNP (`window_size`), the window size in kilobases (`kb_window_size`),
         and the window size in centi Morgan (`cm_window_size`).

         :return: A 2xM matrix of LD boundaries.
        """

        if self.ld_boundaries is None:

            bounds = []

            m = self.genotype_matrix.n_snps
            indices = np.arange(m)

            if self.window_size is not None:
                bounds.append(
                    np.clip(np.array(
                        [indices - self.window_size,
                         indices + self.window_size
                         ]
                    ),  a_min=0, a_max=m)
                )

            from .c_utils import find_windowed_ld_boundaries

            if self.kb_window_size is not None:
                bounds.append(
                    find_windowed_ld_boundaries(.001*self.genotype_matrix.bp_pos,
                                                self.kb_window_size)
                )

            if self.cm_window_size is not None:
                bounds.append(
                    find_windowed_ld_boundaries(self.genotype_matrix.cm_pos,
                                                self.cm_window_size)
                )

            if len(bounds) == 1:
                self.ld_boundaries = bounds[0]
            else:
                self.ld_boundaries = np.array([
                    np.maximum.reduce([b[0, :] for b in bounds]),
                    np.minimum.reduce([b[1, :] for b in bounds])
                ])

        return self.ld_boundaries

    def compute(self,
                output_dir,
                overwrite=True,
                delete_original=True,
                dtype='int16',
                compressor_name='zstd',
                compression_level=7,
                compute_spectral_properties=False) -> LDMatrix:
        """

        Compute the windowed LD matrix and store in Zarr array format.

        :param output_dir: The path where to store the resulting LD matrix.
        :param overwrite: If True, overwrite any existing LD matrices in `temp_dir` and `output_dir`.
        :param delete_original: If True, deletes dense or intermediate LD matrices generated along the way.
        :param dtype: The data type for the entries of the LD matrix.
        :param compressor_name: The name of the compressor to use for the LD matrix.
        :param compression_level: The compression level to use for the LD matrix (1-9).
        :param compute_spectral_properties: If True, compute and store information about the eigenvalues of
        the LD matrix.

        :return: An instance of `LDMatrix` encapsulating the computed LD matrix, its attributes, and metadata.
        """

        ld_mat = super().compute(output_dir,
                                 overwrite=overwrite,
                                 delete_original=delete_original,
                                 compute_spectral_properties=compute_spectral_properties,
                                 dtype=dtype,
                                 compressor_name=compressor_name,
                                 compression_level=compression_level)

        ld_mat.set_store_attr('LD estimator', 'windowed')

        w_properties = {}
        if self.window_size is not None:
            w_properties['Window size'] = self.window_size

        if self.kb_window_size is not None:
            w_properties['Window size (kb)'] = self.kb_window_size

        if self.cm_window_size is not None:
            w_properties['Window size (cM)'] = self.cm_window_size

        ld_mat.set_store_attr('Estimator properties', w_properties)

        if compute_spectral_properties:

            if 'Spectral properties' in ld_mat.list_store_attributes():
                spectral_prop = ld_mat.get_store_attr('Spectral properties')
            else:
                spectral_prop = {
                    'Extremal': ld_mat.estimate_extremal_eigenvalues()
                }

            # Estimate extremal eigenvalues within blocks:
            # To quantify the impact of sparsification, we increase the window sizes here by 20%:

            if self.window_size is not None:
                eig_window_size = int(1.2*self.window_size)
            else:
                eig_window_size = None

            if self.kb_window_size is not None:
                eig_kb_window_size = 1.2*self.kb_window_size
            else:
                eig_kb_window_size = None

            if self.cm_window_size is not None:
                eig_cm_window_size = 1.2*self.cm_window_size
            else:
                eig_cm_window_size = None

            eigs, block_bounds = ld_mat.estimate_extremal_eigenvalues(
                block_size=eig_window_size,
                block_size_kb=eig_kb_window_size,
                block_size_cm=eig_cm_window_size,
                return_block_boundaries=True
            )

            spectral_prop['Eigenvalues per block'] = {**eigs, **block_bounds}

            # Estimate minimum eigenvalues while excluding long-range LD regions (if they're present):
            n_snps_before = ld_mat.n_snps
            ld_mat.filter_long_range_ld_regions()
            n_snps_after = ld_mat.n_snps

            if n_snps_after < n_snps_before:
                spectral_prop['Extremal (excluding LRLD)'] = ld_mat.estimate_extremal_eigenvalues()

            # Update or set the spectral properties attribute:
            ld_mat.set_store_attr('Spectral properties', spectral_prop)

            # Reset the mask:
            ld_mat.reset_mask()

        return ld_mat


class ShrinkageLD(SampleLD):
    """
    A wrapper class to facilitate computing shrinkage-based Linkage-Disequilibrium (LD) matrices.
    Shrinkage LD matrices are a way to reduce noise in the LD matrix by shrinking the off-diagonal pairwise
    correlation coefficients towards zero. This is useful for reducing the noise in the LD matrix and
    improving the quality of downstream analyses.

    The shrinkage estimator implemented uses the shrinking procedure derived in:

    Wen X, Stephens M. USING LINEAR PREDICTORS TO IMPUTE ALLELE FREQUENCIES FROM SUMMARY OR POOLED GENOTYPE DATA.
    Ann Appl Stat. 2010 Sep;4(3):1158-1182. doi: 10.1214/10-aoas338. PMID: 21479081; PMCID: PMC3072818.

    Computing the shrinkage intensity requires specifying the effective population size (Ne) and the sample size
    used to infer the genetic map. In addition, it requires specifying a threshold below which the LD is set to zero.

    !!! note
        The threshold may be adjusted depending on the requested storage data type.

    !!! seealso "See Also"
        * [WindowedLD][magenpy.stats.ld.estimator.WindowedLD]
        * [BlockLD][magenpy.stats.ld.estimator.BlockLD]

    :ivar genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
    :ivar ld_boundaries: The LD boundaries for each variant in the LD matrix.
    :ivar genetic_map_ne: The effective population size (Ne) from which the genetic map is derived.
    :ivar genetic_map_sample_size: The sample size used to infer the genetic map.
    :ivar threshold: The shrinkage cutoff below which the LD is set to zero.

    """

    estimator_id = 'shrinkage'

    def __init__(self,
                 genotype_matrix,
                 genetic_map_ne,
                 genetic_map_sample_size,
                 threshold=1e-3):
        """

        Initialize the shrinkage LD estimator with a genotype matrix and shrinkage parameters.

        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param genetic_map_ne: The effective population size (Ne) from which the genetic map is derived.
        :param genetic_map_sample_size: The sample size used to infer the genetic map.
        :param threshold: The shrinkage cutoff below which the LD is set to zero.
        """

        super().__init__(genotype_matrix=genotype_matrix)

        self.genetic_map_ne = genetic_map_ne
        self.genetic_map_sample_size = genetic_map_sample_size
        self.threshold = threshold

    def compute_ld_boundaries(self):
        """
        Compute the shrinkage-based Linkage-Disequilibrium (LD) boundaries.

        :return: A 2xM matrix of LD boundaries.
        """

        if self.ld_boundaries is None:
            from .c_utils import find_shrinkage_ld_boundaries
            self.ld_boundaries = find_shrinkage_ld_boundaries(self.genotype_matrix.cm_pos,
                                                              self.genetic_map_ne,
                                                              self.genetic_map_sample_size,
                                                              self.threshold)

        return self.ld_boundaries

    def compute(self,
                output_dir,
                overwrite=True,
                delete_original=True,
                dtype='int16',
                compressor_name='zstd',
                compression_level=7,
                compute_spectral_properties=False,
                chunk_size=1000) -> LDMatrix:
        """

        !!! note
            The threshold is adjusted depending on the requested storage data type.

        !!! note
            LD Scores are computed before applying shrinkage.

        :param output_dir: The path where to store the resulting LD matrix.
        :param overwrite: If True, overwrite any existing LD matrices in `temp_dir` and `output_dir`.
        :param delete_original: If True, deletes dense or intermediate LD matrices generated along the way.
        :param dtype: The data type for the entries of the LD matrix.
        :param compressor_name: The name of the compressor to use for the LD matrix.
        :param compression_level: The compression level to use for the LD matrix (1-9).
        :param compute_spectral_properties: If True, compute and store information about the eigenvalues of
        the LD matrix.
        :param chunk_size: An optional parameter that sets the maximum number of rows processed simultaneously.
        The smaller the `chunk_size`, the less memory requirements needed for the shrinkage step.

        :return: An instance of `LDMatrix` encapsulating the computed LD matrix, its attributes, and metadata.

        """

        # Adjust the threshold depending on the requested storage data type:
        if np.issubdtype(dtype, np.integer):
            threshold = max(self.threshold, 1./np.iinfo(dtype).max)
        else:
            threshold = self.threshold

        ld_mat = super().compute(output_dir,
                                 overwrite=overwrite,
                                 delete_original=delete_original,
                                 compute_spectral_properties=False,  # Compute after shrinkage if requested
                                 dtype=dtype,
                                 compressor_name=compressor_name,
                                 compression_level=compression_level)

        from .utils import shrink_ld_matrix

        ld_mat = shrink_ld_matrix(ld_mat,
                                  self.genotype_matrix.cm_pos,
                                  self.genotype_matrix.maf_var,
                                  self.genetic_map_ne,
                                  self.genetic_map_sample_size,
                                  threshold,
                                  chunk_size=chunk_size)

        ld_mat.set_store_attr('LD estimator', 'shrinkage')

        ld_mat.set_store_attr('Estimator properties', {
                    'Genetic map Ne': self.genetic_map_ne,
                    'Genetic map sample size': self.genetic_map_sample_size,
                    'Threshold': threshold
                })

        if compute_spectral_properties:

            spectral_prop = {
                'Extremal': ld_mat.estimate_extremal_eigenvalues()
            }

            cm_ld_bounds_start = self.genotype_matrix.cm_pos[self.ld_boundaries[0, :]]
            cm_ld_bounds_end = self.genotype_matrix.cm_pos[self.ld_boundaries[1, :] - 1]

            median_dist_cm = np.median(cm_ld_bounds_end - cm_ld_bounds_start)

            eigs, block_bounds = ld_mat.estimate_extremal_eigenvalues(block_size_cm=median_dist_cm,
                                                                      return_block_boundaries=True)
            spectral_prop['Eigenvalues per block'] = {**eigs, **block_bounds}

            ld_mat.set_store_attr('Spectral properties', spectral_prop)

        return ld_mat


class BlockLD(SampleLD):
    """
    A wrapper class to facilitate computing block-based Linkage-Disequilibrium (LD) matrices.
    Block-based LD matrices are a way to reduce the memory requirements of the LD matrix by
    computing the pairwise correlation coefficients only between SNPs that are within the same LD block.

    LD blocks can be inferred by external software tools, such as `LDetect` of Berisa and Pickrell (2016):

    Berisa T, Pickrell JK. Approximately independent linkage disequilibrium blocks in human populations.
    Bioinformatics. 2016 Jan 15;32(2):283-5. doi: 10.1093/bioinformatics/btv546.
    Epub 2015 Sep 22. PMID: 26395773; PMCID: PMC4731402.

    The `BlockLD` estimator requires the LD blocks to be provided as input. The LD blocks are a Bx2 matrix
    where B is the number of blocks and the columns are the start and end of each block, respectively.

    !!! seealso "See Also"
        * [WindowedLD][magenpy.stats.ld.estimator.WindowedLD]
        * [ShrinkageLD][magenpy.stats.ld.estimator.ShrinkageLD]

    :ivar genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
    :ivar ld_boundaries: The LD boundaries for each variant in the LD matrix.
    :ivar ld_blocks: The LD blocks, a Bx2 matrix where B is the number of blocks and the columns are
    the start and end of each block, respectively.

    """

    estimator_id = 'block'

    def __init__(self,
                 genotype_matrix,
                 ld_blocks=None,
                 ld_blocks_file=None):
        """
        Initialize the block-based LD estimator with a genotype matrix and LD blocks.

        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param ld_blocks: The LD blocks, a Bx2 matrix where B is the number of blocks and the
        columns are the start and end of each block in units of base pair, respectively.
        :param ld_blocks_file: The path to the LD blocks file
        """

        assert ld_blocks_file is not None or ld_blocks is not None

        super().__init__(genotype_matrix=genotype_matrix)

        if ld_blocks is None:
            from ...parsers.misc_parsers import parse_ld_block_data
            self.ld_blocks = parse_ld_block_data(ld_blocks_file)[self.genotype_matrix.chromosome]
        else:
            self.ld_blocks = ld_blocks

        from ...utils.model_utils import map_variants_to_genomic_blocks

        # Map variants to their associated genomic blocks:
        variants_to_blocks = map_variants_to_genomic_blocks(
            pd.DataFrame({'SNP': self.genotype_matrix.snps,
                          'bp_pos': self.genotype_matrix.bp_pos}),
            pd.DataFrame(self.ld_blocks, columns=['block_start', 'block_end']).assign(
                group=np.arange(len(self.ld_blocks))),
            variant_pos_col='bp_pos',
            filter_unmatched=True
        )

        # Split the genotype matrix by the blocks:
        split_geno_matrices = self.genotype_matrix.split_by_variants(variants_to_blocks)

        self._block_estimators = {
            i: SampleLD(geno_matrix) for i, geno_matrix in split_geno_matrices.items()
        }

    def compute_ld_boundaries(self):
        """
        Compute the per-SNP Linkage-Disequilibrium (LD) boundaries for the block-based estimator.

        :return: A 2xM matrix of LD boundaries.
        """

        if self.ld_boundaries is None:
            from .c_utils import find_ld_block_boundaries
            self.ld_boundaries = find_ld_block_boundaries(self.genotype_matrix.bp_pos, self.ld_blocks)

        return self.ld_boundaries

    def compute(self,
                output_dir,
                overwrite=True,
                delete_original=True,
                dtype='int16',
                compressor_name='zstd',
                compression_level=7,
                compute_spectral_properties=False) -> LDMatrix:
        """

        Compute the block-based LD matrix and store in Zarr array format.

        :param output_dir: The path where to store the resulting LD matrix.
        :param overwrite: If True, overwrite any existing LD matrices in `temp_dir` and `output_dir`.
        :param delete_original: If True, deletes dense or intermediate LD matrices generated along the way.
        :param compute_spectral_properties: If True, compute and store information about the eigenvalues of
        the LD matrix.
        :param dtype: The data type for the entries of the LD matrix.
        :param compressor_name: The name of the compressor to use for the LD matrix.
        :param compression_level: The compression level to use for the LD matrix (1-9).

        :return: An instance of `LDMatrix` encapsulating the computed LD matrix, its attributes, and metadata.
        """

        ld_mats = []

        # TODO: Parallelize this loop

        temp_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix=self.temp_dir_prefix)

        for i, block_estimator in self._block_estimators.items():

            block_output_dir = osp.join(temp_dir.name,
                                        f'chr_{self.genotype_matrix.chromosome}_block_{i}/')

            makedir(block_output_dir)

            ld_mats.append(
                block_estimator.compute(
                    block_output_dir,
                    overwrite=overwrite,
                    delete_original=delete_original,
                    dtype=dtype,
                    compressor_name=compressor_name,
                    compression_level=compression_level,
                    compute_spectral_properties=compute_spectral_properties
                )
            )

        # If the user requested computing the spectral properties, we need to obtain
        # the minimum eigenvalue from the blocks:
        if compute_spectral_properties:

            extremal_eigs = pd.DataFrame([ld.get_store_attr('Spectral properties')['Extremal']
                                          for ld in ld_mats])
            blocks = np.insert(np.cumsum([ld.stored_n_snps for ld in ld_mats]), 0, 0)
            block_starts = blocks[:-1]
            block_ends = blocks[1:]

        # Combine the LD matrices for the individual blocks into a single LD matrix:
        from .utils import combine_ld_matrices

        output_dir = osp.join(output_dir, f'chr_{self.genotype_matrix.chromosome}/')

        ld_mat = combine_ld_matrices(ld_mats,
                                     output_dir,
                                     overwrite=overwrite,
                                     delete_original=delete_original)

        # Populate the attributes of the LDMatrix object:
        ld_mat.set_store_attr('LD estimator', 'block')
        ld_mat.set_store_attr('Estimator properties', {
            'LD blocks': self.ld_blocks.tolist()
        })

        ld_mat.set_store_attr('Chromosome', int(self.genotype_matrix.chromosome))
        ld_mat.set_store_attr('Sample size', int(self.genotype_matrix.sample_size))

        if self.genotype_matrix.genome_build is not None:
            ld_mat.set_store_attr('Genome build', self.genotype_matrix.genome_build)

        if compute_spectral_properties:

            spectral_prop = {
                'Extremal': {
                    'min': extremal_eigs['min'].min(),
                    'max': extremal_eigs['max'].max()
                },
                'Eigenvalues per block': {
                    'min': list(extremal_eigs['min']),
                    'max': list(extremal_eigs['max']),
                    'block_start': list(block_starts),
                    'block_end': list(block_ends)
                }
            }

            ld_mat.set_store_attr('Spectral properties', spectral_prop)

        if ld_mat.validate_ld_matrix():
            temp_dir.cleanup()
            return ld_mat




import numpy as np


class AnnotationMatrix(object):
    """
    A wrapper class for handling annotation matrices, which are essentially tables of
    features for each variant in the genome. These features include information such as
    whether the variant is in coding regions, enhancers, etc. It can also include continuous
    features derived from experimental assays or other sources.

    The purpose of this class is to present a unified and consistent interface for handling
    annotations across different tools and applications. It should be able to read and write
    annotation matrices in different formats, filter annotations, and perform basic operations
    on the annotation matrix. It should also allow users to define new custom annotations
    that can be used for downstream statistical genetics applications.

    :ivar table: A pandas dataframe containing the annotation information.
    :ivar _annotations: A list or array of column namees to consider as annotations. If not provided,
    will be inferred heuristically, though we recommend that the user specify this information.
    """

    def __init__(self, annotation_table=None, annotations=None):
        """
        Initialize an AnnotationMatrix object

        :param annotation_table: A pandas dataframe containing the annotation information.
        :param annotations: A list of array of columns to consider as annotations. If not provided, will be
        inferred heuristically, though we recommend that the user specify this information.
        """

        self.table = annotation_table
        self._annotations = annotations

        if self.table is not None:
            if self._annotations is None:
                self._annotations = [ann for ann in self.table.columns if ann not in ('CHR', 'SNP', 'POS')]
                if len(self._annotations) < 1:
                    self._annotations = None

    @classmethod
    def from_file(cls, annot_file, annot_format='magenpy', annot_parser=None,
                  **parse_kwargs):
        """
        Initialize an AnnotationMatrix object from a file.

        :param annot_file: The path to the annotation file.
        :param annot_format: The format of the annotation file. For now, we mainly support
        annotation files in the `magenpy` and `ldsc` formats.
        :param annot_parser: An `AnnotationMatrixParser` derived object, which can be tailored to
        specific annotation formats that the user has.
        :param parse_kwargs: arguments for the pandas `read_csv` function, such as the delimiter.

        :return: An instance of the `AnnotationMatrix` class.
        """

        from .parsers.annotation_parsers import AnnotationMatrixParser, LDSCAnnotationMatrixParser

        if annot_parser is None:
            if annot_format == 'magenpy':
                annot_parser = AnnotationMatrixParser(None, **parse_kwargs)
            elif annot_format == 'ldsc':
                annot_parser = LDSCAnnotationMatrixParser(None, **parse_kwargs)
            else:
                raise KeyError(f"Annotation matrix format {annot_format} not recognized!")

        annot_table, annotations = annot_parser.parse(annot_file)

        annot_mat = cls(annotation_table=annot_table, annotations=annotations)

        return annot_mat

    @property
    def shape(self):
        """
        :return: The dimensions of the annotation matrix (number of variants x number of annotations).
        """
        return self.n_snps, self.n_annotations

    @property
    def n_snps(self):
        """
        :return: The number of variants in the annotation matrix.
        """
        return len(self.table)

    @property
    def chromosome(self):
        """
        A convenience method to get the chromosome if there is only one chromosome in the annotation matrix.

        :return: The chromosome number if there is only one chromosome in the annotation matrix. Otherwise, None.
        """
        chrom = self.chromosomes
        if chrom is not None:
            if len(chrom) == 1:
                return chrom[0]

    @property
    def chromosomes(self):
        """
        :return: The list of unique chromosomes in the annotation matrix.
        """
        if 'CHR' in self.table.columns:
            return self.table['CHR'].unique()

    @property
    def snps(self):
        """
        :return: The list of SNP rsIDs in the annotation matrix.
        """
        return self.table['SNP'].values

    @property
    def n_annotations(self):
        """
        :return: The number of annotations in the annotation matrix.
        """
        if self.annotations is None:
            return 0
        else:
            return len(self.annotations)

    @property
    def binary_annotations(self):
        """
        :return: A list of binary (0/1) annotations in the annotation matrix.
        """
        assert self.annotations is not None
        return np.array([c for c in self.annotations
                         if len(self.table[c].unique()) == 2])

    @property
    def annotations(self):
        """
        :return: The list of annotation names or IDs in the annotation matrix.
        """
        return self._annotations

    def values(self, add_intercept=False):
        """
        :param add_intercept: Adds a base annotation corresponding to the intercept.

        :return: The annotation matrix as a numpy matrix.
        :raises KeyError: If no annotations are defined in the table.
        """

        if self.annotations is None:
            raise KeyError("No annotations are defined in this table!")
        annot_mat = self.table[self.annotations].values
        if add_intercept:
            return np.hstack([np.ones((annot_mat.shape[0], 1)), annot_mat])
        else:
            return annot_mat

    def filter_snps(self, extract_snps=None, extract_file=None):
        """
        Filter variants from the annotation matrix. User must specify
        either a list of variants to extract or the path to a file
        with the list of variants to extract.

        :param extract_snps: A list or array of SNP IDs to keep in the annotation matrix.
        :param extract_file: The path to a file with the list of variants to extract.
        """

        assert extract_snps is not None or extract_file is not None

        if extract_file is not None:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        from .utils.compute_utils import intersect_arrays

        arr_idx = intersect_arrays(self.snps, extract_snps, return_index=True)

        self.table = self.table.iloc[arr_idx, :].reset_index()

    def filter_annotations(self, keep_annotations):
        """
        Filter the list of annotations in the matrix.
        :param keep_annotations: A list or array of annotations to keep.
        """

        if self.annotations is None:
            return

        self._annotations = [annot for annot in self._annotations if annot in keep_annotations]
        self.table = self.table[['CHR', 'SNP', 'POS'] + self._annotations]

    def add_annotation(self, annot_vec, annotation_name):
        """
        Add an annotation vector or list to the AnnotationMatrix object.

        :param annot_vec: A vector/list/Series containing the annotation information for each SNP in the
        AnnotationMatrix. For now, it's the responsibility of the user to make sure that the annotation
        list or vector are sorted properly.
        :param annotation_name: The name of the annotation to create. Make sure the name is not already
        in the matrix!
        """

        if self.annotations is not None:
            assert annotation_name not in self.annotations
        assert len(annot_vec) == self.n_snps

        self.table[annotation_name] = annot_vec

        if self.annotations is None:
            self._annotations = [annotation_name]
        else:
            self._annotations = list(self._annotations) + [annotation_name]

    def add_annotation_from_bed(self, bed_file, annotation_name):
        """
        Add an annotation to the AnnotationMatrix from a BED file that lists
        the range of coordinates associated with that annotation (e.g. coding regions, enhancers, etc.).
        The BED file has to adhere to the format specified by,
        https://uswest.ensembl.org/info/website/upload/bed.html
        with the first three columns being:

        CHR StartCoordinate EndCoordinate ...

        !!! note
            This implementation is quite slow at the moment. May need to find more efficient
            ways to do the merge over list of ranges.

        :param bed_file: The path to the BED file containing the annotation coordinates.
        :param annotation_name: The name of the annotation to create. Make sure the name is not already
        in the matrix!

        :raises AssertionError: If the annotation name is already in the matrix.
        """

        from .parsers.annotation_parsers import parse_annotation_bed_file

        if self.annotations is not None:
            assert annotation_name not in self.annotations

        bed_df = parse_annotation_bed_file(bed_file)
        # Group the BED annotation file by chromosome:
        range_groups = bed_df.groupby('CHR').groups

        def annotation_overlap(row):
            """
            This function takes a row from the annotation matrix table
            and returns True if and only if the BP position for the
            SNP is within the range specified by the annotation BED file.
            """
            try:
                chr_range = bed_df.iloc[range_groups[row['CHR']], :]
            except KeyError:
                return False

            check = (chr_range.Start <= row['POS']) & (chr_range.End >= row['POS'])
            return int(np.any(check))

        self.table[annotation_name] = self.table.apply(annotation_overlap, axis=1)

        if self.annotations is None:
            self._annotations = [annotation_name]
        else:
            self._annotations = list(self._annotations) + [annotation_name]

    def get_binary_annotation_index(self, bin_annot):
        """
        :param bin_annot: The name of the binary annotation for which to fetch the relevant variants.
        :return: The indices of all variants that belong to binary annotation `bin_annot`
        """
        assert bin_annot in self.binary_annotations
        return np.where(self.table[bin_annot] == 1)[0]

    def split_by_chromosome(self):
        """
        Split the annotation matrix by chromosome.

        :return: A dictionary of `AnnotationMatrix` objects, where the keys are the chromosome numbers.
        """

        if 'CHR' in self.table.columns:
            chrom_tables = self.table.groupby('CHR')
            return {
                c: AnnotationMatrix(annotation_table=chrom_tables.get_group(c),
                                    annotations=self.annotations)
                for c in chrom_tables.groups
            }
        else:
            raise KeyError("Chromosome information is not available in the annotation table!")

    def to_file(self, output_path, col_subset=None, compress=True, **to_csv_kwargs):
        """
        A convenience method to write the annotation matrix to a file.

        :param output_path: The path and prefix to the file where to write the annotation matrix.
        :param col_subset: A subset of the columns to write to file.
        :param compress: Whether to compress the output file (default: True).
        :param to_csv_kwargs: Key-word arguments to the pandas csv writer.
        """

        if 'sep' not in to_csv_kwargs and 'delimiter' not in to_csv_kwargs:
            to_csv_kwargs['sep'] = '\t'

        if 'index' not in to_csv_kwargs:
            to_csv_kwargs['index'] = False

        if col_subset is not None:
            table = self.table[col_subset]
        else:
            table = self.table

        file_name = output_path + '.annot'
        if compress:
            file_name += '.gz'

        table.to_csv(file_name, **to_csv_kwargs)



import numpy as np
import magenpy as mgp
import shutil
import pytest


def _basic_ld_checks(gdl: mgp.GWADataLoader):
    """
    Perform basic checks on the computed  LD matrix, such as its dimensions,
    sample size, and the number of SNPs.

    :param gdl: An instance of `GWADataLoader`

    :return: True if all checks pass, raises AssertionError otherwise.
    """

    assert gdl.ld is not None
    assert gdl.ld[22] is not None
    assert gdl.ld[22].validate_ld_matrix()
    assert gdl.ld[22].stored_n_snps == gdl.n_snps
    assert gdl.ld[22].sample_size == gdl.sample_size

    assert np.array_equal(gdl.ld[22].snps, gdl.snps[22])
    # Add other checks?
    # ...

    return True


@pytest.fixture(scope='module')
def gdl_object():
    """
    Initialize a GWADataLoader using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    gdl = mgp.GWADataLoader(mgp.tgp_eur_data_path(),
                            backend='xarray')

    # Extract a smaller subset of variants for testing:
    np.random.seed(0)
    keep_snps = gdl.snps[22][np.random.choice(2000, 1000, replace=False)]
    gdl.filter_snps(extract_snps=keep_snps, chromosome=22)

    yield gdl

    # Clean up after tests are done:
    gdl.cleanup()
    shutil.rmtree(gdl.temp_dir)
    shutil.rmtree(gdl.output_dir)


@pytest.fixture(scope='module')
def sample_ld(gdl_object: mgp.GWADataLoader):
    """
    Test the LD computation functionality according to the Sample estimator
    """

    gdl_object.compute_ld('sample', gdl_object.output_dir)
    gdl_object.harmonize_data()

    _basic_ld_checks(gdl_object)

    yield gdl_object.ld[22]


@pytest.fixture(scope='module')
def windowed_ld(gdl_object: mgp.GWADataLoader):
    """
    Test the LD computation functionality according to the Windowed estimator
    """

    gdl_object.compute_ld('windowed',
                          gdl_object.output_dir,
                          compute_spectral_properties=True,
                          window_size=500,
                          kb_window_size=100,
                          cm_window_size=3.)
    gdl_object.harmonize_data()

    _basic_ld_checks(gdl_object)

    yield gdl_object.ld[22]


@pytest.fixture(scope='module')
def shrinkage_ld(gdl_object: mgp.GWADataLoader):
    """
    Test the LD computation functionality according to the Shrinkage estimator
    """

    gdl_object.compute_ld('shrinkage',
                          gdl_object.output_dir,
                          genetic_map_ne=11400,
                          genetic_map_sample_size=183)
    gdl_object.harmonize_data()

    _basic_ld_checks(gdl_object)

    yield gdl_object.ld[22]


@pytest.fixture(scope='module')
def block_ld(gdl_object: mgp.GWADataLoader):
    """
    Test the LD computation functionality according to the Block estimator
    """

    ld_block_url = ("https://bitbucket.org/nygcresearch/ldetect-data/raw/"
                    "ac125e47bf7ff3e90be31f278a7b6a61daaba0dc/EUR/fourier_ls-all.bed")
    gdl_object.compute_ld('block', gdl_object.output_dir, ld_blocks_file=ld_block_url)
    gdl_object.harmonize_data()

    _basic_ld_checks(gdl_object)

    yield gdl_object.ld[22]


def test_basic_ld_filtering(block_ld: mgp.LDMatrix):
    """
    Test the LD filtering functionality
    """

    # First release the data from memory and reset the mask:
    block_ld.reset_mask()
    block_ld.release()

    orig_n_snps = block_ld.n_snps

    # Sample a subset of SNPs:
    snp_subset = np.random.choice(block_ld.snps, 100, replace=False)
    block_ld.filter_snps(snp_subset)

    # Check that the number of SNPs is as expected:
    assert block_ld.n_snps == len(snp_subset)
    assert block_ld.shape == (len(snp_subset), len(snp_subset))

    intersected_snps = np.intersect1d(block_ld.snps, snp_subset)

    assert len(intersected_snps) == len(snp_subset)
    assert np.all(np.isin(snp_subset, block_ld.snps))

    # Check the snp table:
    assert block_ld.to_snp_table().shape[0] == len(snp_subset)

    # Load a linear operator object:
    lop = block_ld.to_linear_operator()

    assert lop.shape == block_ld.shape

    # check multiplication:
    x = np.random.rand(block_ld.n_snps)
    assert block_ld.dot(x).shape == x.shape

    # Check that attributes regarding the original data are still correct:
    assert block_ld.stored_n_snps == orig_n_snps
    assert block_ld.stored_shape == (orig_n_snps, orig_n_snps)


def test_ld_linear_operator(block_ld: mgp.LDMatrix):
    """
    Test the Linear Operator functionality
    """

    # First release the data from memory and reset the mask:
    block_ld.reset_mask()
    block_ld.release()

    # Load a linear operator object:
    lop = block_ld.to_linear_operator()

    assert lop.shape == block_ld.shape

    # check multiplication:
    x = np.random.rand(block_ld.n_snps)
    lop_dot = lop.dot(x)
    assert lop_dot.shape == x.shape

    # Check the data type:
    assert block_ld.dtype == lop.ld_data_type

    # Check CSR functionality:
    csr_mat = lop.to_csr()
    assert csr_mat.shape == lop.shape

    # Check that multiplication with CSR yields the same result:
    csr_dot = csr_mat.dot(x)
    assert np.allclose(csr_dot, lop_dot)

    # Set up different boundaries for testing slicing functionality:
    boundaries = [(0, 100), (150, 250), (lop.shape[0] - 100, lop.shape[0])]

    for start, end in boundaries:

        # Extract a numpy block from the linear operator directly:
        numpy_block = lop.to_numpy(block_start=start, block_end=end)

        # Extract the same block from the CSR matrix:
        csr_block = csr_mat[start:end, start:end].toarray()

        # Slice the LOP object:
        lop_block = lop[start:end, start:end]

        # Check that all the resulting shapes are compatible:
        assert numpy_block.shape == csr_block.shape == lop_block.shape

        # Load the lop block into a numpy array:
        lop_block = lop_block.to_numpy()

        assert np.allclose(lop_block, csr_block) and np.allclose(numpy_block, csr_block)

        # Check that the block is symmetric:
        assert np.allclose(numpy_block, numpy_block.T)

        # Check that the block contains all ones along the diagonal:
        assert np.allclose(np.diag(numpy_block), 1.)




import os.path as osp


def tgp_eur_data_path():
    """
    :return: The path of the attached 1000G genotype data for
    European samples (N=378) and a subset of chromosome 22 (p=15938)
    """
    return osp.join(osp.dirname(osp.dirname(__file__)), 'data/1000G_eur_chr22')


def ukb_height_sumstats_path():
    """
    :return: The path of the attached GWAS summary statistics file
    for standing height. The file contains summary statistics for
    HapMap3 variants on CHR22 and is a snapshot of the summary statistics
    published on the fastGWA database:
    https://yanglab.westlake.edu.cn/data/fastgwa_data/UKB/50.v1.1.fastGWA.gz
    """
    return osp.join(osp.dirname(osp.dirname(__file__)), 'data/ukb_height_chr22.fastGWA.gz')


def lrld_path():
    """
    The boundaries of Long Range LD (LRLD) regions derived from here:

        https://genome.sph.umich.edu/wiki/Regions_of_high_linkage_disequilibrium_(LD)

    Which is based on the work of

    > Anderson, Carl A., et al. "Data quality control in genetic case-control association studies."
    Nature protocols 5.9 (2010): 1564-1573.

    :return: The path of the attached BED file containing long-range linkage disequilibrium
    (LD) regions in the human genome. The coordinates are in hg19/GRCh37.
    """
    return osp.join(osp.dirname(osp.dirname(__file__)), 'data/lrld_hg19_GRCh37.txt')



# cython: linetrace=False
# cython: profile=False
# cython: binding=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: language_level=3
# cython: infer_types=True


from libc.math cimport exp
from libc.stdint cimport int64_t, int32_t
from cython cimport floating
cimport cython
import numpy as np
cimport numpy as cnp


ctypedef fused int_dtype:
    cnp.int32_t
    cnp.int64_t

ctypedef fused noncomplex_numeric:
    cnp.int8_t
    cnp.int16_t
    cnp.int32_t
    cnp.int64_t
    cnp.float32_t
    cnp.float64_t


cdef extern from "ld_utils.hpp" nogil:
    bint blas_supported() noexcept nogil
    bint omp_supported() noexcept nogil

    void ld_matrix_dot[T, U, I](int c_size,
              int* ld_left_bound,
              I* ld_indptr,
              U* ld_data,
              T* vec,
              T* out,
              T dq_scale,
              int threads) noexcept nogil

    void ut_ld_matrix_dot[T, U, I](int c_size,
              I* ld_indptr,
              U* ld_data,
              T* vec,
              T* out,
              T dq_scale,
              bint include_lower_triangle,
              bint include_upper_triangle,
              bint include_diag,
              int threads) noexcept nogil

    void ld_rank_one_update[T, U, I](int c_size,
                   int* ld_left_bound,
                   I* ld_indptr,
                   U* ld_data,
                   T* vec,
                   T* out,
                   T alpha,
                   T dq_scale,
                   int threads) noexcept nogil


cpdef ld_dot(int[::1] ld_left_bound,
             int_dtype[::1] ld_indptr,
             noncomplex_numeric[::1] ld_data,
             floating[::1] vec,
             floating dq_scale = 1.,
             int threads = 1):

    cdef:
        floating[::1] out = np.zeros_like(vec)

    ld_matrix_dot(vec.shape[0],
                  &ld_left_bound[0],
                  &ld_indptr[0],
                  &ld_data[0],
                  &vec[0],
                  &out[0],
                  dq_scale,
                  threads)

    return np.asarray(out)


cpdef ut_ld_dot(int_dtype[::1] ld_indptr,
             noncomplex_numeric[::1] ld_data,
             floating[::1] vec,
             floating dq_scale = 1.,
             bint include_lower_triangle = 1,
             bint include_upper_triangle = 1,
             bint include_diag = 1,
             int threads = 1):

    cdef:
        floating[::1] out = np.zeros_like(vec)

    ut_ld_matrix_dot(vec.shape[0],
                  &ld_indptr[0],
                  &ld_data[0],
                  &vec[0],
                  &out[0],
                  dq_scale,
                  include_lower_triangle,
                  include_upper_triangle,
                  include_diag,
                  threads)

    return np.asarray(out)


cpdef rank_one_update(int[::1] ld_left_bound,
                      int_dtype[::1] ld_indptr,
                      noncomplex_numeric[::1] ld_data,
                      floating[::1] vec,
                      floating[::1] out,
                      floating alpha = 1.,
                      floating dq_scale = 1.,
                      int threads = 1):

    ld_rank_one_update(
        vec.shape[0],
        &ld_left_bound[0],
        &ld_indptr[0],
        &ld_data[0],
        &vec[0],
        &out[0],
        alpha,
        dq_scale,
        threads
    )

# -------------------------------------------------------------------------
# Utilities for manipulating/interacting with generic numeric types:

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef noncomplex_numeric numeric_abs(noncomplex_numeric x) noexcept nogil:
    """
    Return the absolute value of a numeric type.
    """
    if x < 0:
        return -x
    return x


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef noncomplex_numeric numeric_max(noncomplex_numeric x, noncomplex_numeric y) noexcept nogil:
    """
    Return maximum of two numeric values.
    """
    if x > y:
        return x
    return y

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef noncomplex_numeric numeric_min(noncomplex_numeric x, noncomplex_numeric y) noexcept nogil:
    """
    Return minimum of two numeric values.
    """
    if x < y:
        return x
    return y

# -------------------------------------------------------------------------
# Utilities for slicing, filtering, symmetrizing and interacting with the LD matrix:

cpdef find_tagging_variants(int[::1] variant_indices,
                           int_dtype[::1] indptr,
                           noncomplex_numeric[::1] data,
                           noncomplex_numeric threshold):
    """
    TODO: Implement function to find tagging variants.
    """
    pass

cpdef prune_ld_ut(int_dtype[::1] indptr,
                  noncomplex_numeric[::1] data,
                  noncomplex_numeric r_threshold):
    """
    Pass over the LD matrix once and prune it so that variants whose absolute correlation coefficient is above 
    or equal to a certain threshold are filtered away. If two variants are highly correlated, 
    this function keeps the one that occurs earlier in the matrix. 
    
    This function works with LD matrices in any data type 
    (quantized to integers or floats), but it is the user's responsibility to set the appropriate 
    threshold for the data type used.
    
    !!! note 
        This function assumes that the LD matrix is in upper triangular form and doesn't include the 
        diagonal. We will try to generalize this implementation later.
    
    :param indptr: The index pointer array for the CSR matrix to be pruned.
    :param data: The data array for the CSR matrix to be pruned.
    :param r_threshold: The Pearson Correlation coefficient threshold above which to prune variants.
    
    :return: An boolean array of which variants are kept after pruning.
    """

    cdef:
        int64_t i, curr_row, curr_row_size, curr_data_idx, curr_shape=indptr.shape[0]-1
        char[::1] keep = np.ones(curr_shape, dtype=np.int8)

    with nogil:
        for curr_row in range(curr_shape):

            if keep[curr_row] == 1:

                curr_row_size = indptr[curr_row + 1] - indptr[curr_row]

                for i in range(curr_row_size):
                    curr_data_idx = indptr[curr_row] + i

                    if numeric_abs(data[curr_data_idx]) >= r_threshold:
                        keep[curr_row + i + 1] = 0

    return np.asarray(keep).view(bool)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef get_symmetrized_indptr_with_mask(int_dtype[::1] indptr,
                                       cnp.ndarray[cnp.npy_bool, ndim=1] mask):
    """
    Given an index pointer array from an upper triangular CSR matrix, this function 
    computes the equivalent indptr for the symmetric matrix and returns also the 
    column index of the leftmost element for each row. This is a utility function 
    mainly used to help symmetrize upper triangular and block-diagonal CSR matrices with minimal 
    memory overhead. The function also supports filtering the matrix by using a boolean mask.
    
    :param indptr: The index pointer array for the CSR matrix to be symmetrized.
    :param mask: A boolean mask indicating which elements (rows) of the matrix to keep.
    
    :return: A tuple with the new indptr array and the leftmost column index for each row.
    """

    # Compute the cumulative number of skipped rows:
    cum_skipped_rows = np.cumsum(~mask, dtype=np.int32)

    # Determine the rightmost element for every row:
    rightmost = np.diff(indptr).astype(np.int32) + np.arange(indptr.shape[0] - 1, dtype=np.int32)
    # Update the index by taking into account the number of skipped rows:
    rightmost -= cum_skipped_rows[rightmost]
    # Keep rows indicated by the mask:
    rightmost = rightmost[mask]

    # Get unique boundaries:
    uniq_res = np.unique(rightmost, return_index=True)

    # Loop over the remaining rows to get the leftmost index:
    cdef:
        int32_t curr_row, shape=rightmost.shape[0]
        int32_t[::1] leftmost = np.zeros(shape, dtype=np.int32)
        int32_t[::1] uniq_rightmost = uniq_res[0]
        int32_t[::1] rightmost_first_idx = uniq_res[1].astype(np.int32)
        int32_t rightmost_idx = 0, uniq_rightmost_size = uniq_res[0].shape[0]

    with nogil:
        for curr_row in range(shape):

            if curr_row > uniq_rightmost[rightmost_idx] and rightmost_idx < uniq_rightmost_size - 1:
                rightmost_idx += 1

            leftmost[curr_row] = rightmost_first_idx[rightmost_idx]


    leftmost = np.asarray(leftmost)

    # Compute the new indptr:
    new_indptr = np.zeros(leftmost.shape[0] + 1, dtype=np.int64)
    new_indptr[1:] += rightmost + 1 - leftmost
    np.cumsum(new_indptr, out=new_indptr)

    return new_indptr, leftmost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef get_symmetrized_indptr(int_dtype[::1] indptr):
    """
    Given an index pointer array from an upper triangular CSR matrix, this function 
    computes the equivalent indptr for the symmetric matrix and returns also the 
    column index of the leftmost element for each row. This is a utility function 
    mainly used to help symmetrize upper triangular and block-diagonal CSR matrices with minimal 
    memory overhead.

    :param indptr: The index pointer array for the CSR matrix to be symmetrized.

    :return: A tuple with the new indptr array and the leftmost column index for each row.
    """

    # Determine the rightmost element for every row:
    rightmost = np.diff(indptr).astype(np.int32) + np.arange(indptr.shape[0] - 1, dtype=np.int32)

    # Get unique boundaries:
    uniq_res = np.unique(rightmost, return_index=True)

    # Loop over the remaining rows to get the leftmost index:
    cdef:
        int32_t curr_row, shape=rightmost.shape[0]
        int32_t[::1] leftmost = np.zeros(shape, dtype=np.int32)
        int32_t[::1] uniq_rightmost = uniq_res[0]
        int32_t[::1] rightmost_first_idx = uniq_res[1].astype(np.int32)
        int32_t rightmost_idx = 0, uniq_rightmost_size = uniq_res[0].shape[0]

    with nogil:
        for curr_row in range(shape):

            if curr_row > uniq_rightmost[rightmost_idx] and rightmost_idx < uniq_rightmost_size - 1:
                rightmost_idx += 1

            leftmost[curr_row] = rightmost_first_idx[rightmost_idx]


    leftmost = np.asarray(leftmost)

    # Compute the new indptr:
    new_indptr = np.zeros(leftmost.shape[0] + 1, dtype=np.int64)
    new_indptr[1:] += rightmost + 1 - leftmost
    np.cumsum(new_indptr, out=new_indptr)

    return new_indptr, leftmost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef symmetrize_ut_csr_matrix_with_mask(int_dtype[::1] indptr,
                                         noncomplex_numeric[::1] data,
                                         cnp.ndarray[cnp.npy_bool, ndim=1] mask,
                                         noncomplex_numeric diag_fill_value):
    """
    Given an upper triangular CSR matrix, this function symmetrizes it by adding the 
    transpose of the upper triangular matrix to itself. This function assumes the following:
    
        1. The non-zero elements are contiguous along the diagonal of each row (starting from 
        the diagonal + 1).
        2. The diagonal elements aren't present in the upper triangular matrix.
        
    The function also supports filtering the matrix by using a boolean mask.
    
    :param indptr: The index pointer array for the CSR matrix to be symmetrized.
    :param data: The data array for the CSR matrix to be symmetrized.
    :param mask: A boolean mask indicating which elements (rows) of the matrix to keep.
    :param diag_fill_value: The value to fill the diagonal with (equivalent to 1 for the various data types).
    
    :return: A tuple with the new data array, the new indptr array, and the leftmost column index for each row.
    
    """

    new_idx = get_symmetrized_indptr_with_mask(indptr, mask)

    cdef:
        int64_t[::1] new_indptr = new_idx[0]
        int32_t[::1] leftmost_col = new_idx[1]
        int32_t[::1] cum_skipped_rows = np.cumsum(~mask, dtype=np.int32)
        int32_t curr_row, curr_col, curr_row_size, filt_curr_row, filt_curr_col, curr_shape=indptr.shape[0]-1
        int64_t curr_data_idx, new_idx_1, new_idx_2
        noncomplex_numeric[::1] new_data = np.empty_like(data, shape=(new_idx[0][new_indptr.shape[0] - 1], ))


    with nogil:

        # For each row in the current matrix:
        for curr_row in range(curr_shape):
            if mask[curr_row]:

                # Determine the row size for the upper triangular matrix:
                curr_row_size = indptr[curr_row + 1] - indptr[curr_row]

                filt_curr_row = curr_row - cum_skipped_rows[curr_row]

                # First, add the identity to the diagonal
                new_idx_1 = new_indptr[filt_curr_row] + filt_curr_row - leftmost_col[filt_curr_row]
                new_data[new_idx_1] = diag_fill_value

                # Then, add and reflect the off-diagonal entries:
                for curr_col in range(curr_row + 1, curr_row_size + curr_row + 1):
                    if mask[curr_col]:

                        filt_curr_col = curr_col - cum_skipped_rows[curr_col]

                        curr_data_idx = indptr[curr_row] + curr_col - curr_row - 1

                        new_idx_1 = new_indptr[filt_curr_row] + filt_curr_col - leftmost_col[filt_curr_row]
                        new_idx_2 = new_indptr[filt_curr_col] + filt_curr_row - leftmost_col[filt_curr_col]

                        new_data[new_idx_1] = data[curr_data_idx]
                        new_data[new_idx_2] = data[curr_data_idx]

    return np.asarray(new_data), np.asarray(new_idx[0]), np.asarray(new_idx[1])

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef symmetrize_ut_csr_matrix(int_dtype[::1] indptr,
                                noncomplex_numeric[::1] data,
                                noncomplex_numeric diag_fill_value):
    """
    Given an upper triangular CSR matrix, this function symmetrizes it by adding the 
    transpose of the upper triangular matrix to itself. This function assumes the following:

        1. The non-zero elements are contiguous along the diagonal of each row (starting from 
        the diagonal + 1).
        2. The diagonal elements aren't present in the upper triangular matrix.

    :param indptr: The index pointer array for the CSR matrix to be symmetrized.
    :param data: The data array for the CSR matrix to be symmetrized.
    :param diag_fill_value: The value to fill the diagonal with (equivalent to 1 for the various data types).

    :return: A tuple with the new data array, the new indptr array, and the leftmost column index for each row.

    """

    new_idx = get_symmetrized_indptr(indptr)

    cdef:
        int64_t[::1] new_indptr = new_idx[0]
        int32_t[::1] leftmost_col = new_idx[1]
        int32_t curr_row, curr_col, curr_row_size, curr_shape=indptr.shape[0] - 1
        int64_t curr_data_idx, new_idx_1, new_idx_2
        noncomplex_numeric[::1] new_data = np.empty_like(data, shape=(new_idx[0][new_indptr.shape[0] - 1], ))

    with nogil:

        # For each row in the current matrix:
        for curr_row in range(curr_shape):

            # Determine the row size for the upper triangular matrix:
            curr_row_size = indptr[curr_row + 1] - indptr[curr_row]

            # First, add the identity to the diagonal
            new_idx_1 = new_indptr[curr_row] + curr_row - leftmost_col[curr_row]
            new_data[new_idx_1] = diag_fill_value

            # Then, add and reflect the off-diagonal entries:
            for curr_col in range(curr_row + 1, curr_row_size + curr_row + 1):

                curr_data_idx = indptr[curr_row] + curr_col - curr_row - 1

                new_idx_1 = new_indptr[curr_row] + curr_col - leftmost_col[curr_row]
                new_data[new_idx_1] = data[curr_data_idx]

                # To allow for non-square matrices, we need to check if the column index is within bounds:
                if curr_col < curr_shape:
                    new_idx_2 = new_indptr[curr_col] + curr_row - leftmost_col[curr_col]
                    new_data[new_idx_2] = data[curr_data_idx]

    return np.asarray(new_data), np.asarray(new_idx[0]), np.asarray(new_idx[1])


cpdef filter_ut_csr_matrix_inplace(int_dtype[::1] indptr,
                                  noncomplex_numeric[::1] data,
                                  char[::1] bool_mask,
                                  int new_size):
    """
    Given an upper triangular CSR matrix represented by the data and indptr arrays, this function filters 
    its entries with a boolean mask.

        1. The non-zero elements are contiguous along the diagonal of each row (starting from 
        the diagonal + 1).
        2. The diagonal elements aren't present in the upper triangular matrix.
        
    !!! warning
        This function modifies the input data array inplace.

    :param indptr: The index pointer array for the CSR matrix to be filtered.
    :param data: The data array for the CSR matrix to be filtered.
    :param bool_mask: A boolean mask indicating which elements (rows) of the matrix to keep.
    :param new_size: The new size of the filtered matrix.

    :return: A tuple with the new filtered data array and the new indptr array.
    """

    cdef:
        int64_t i, curr_row, row_bound, new_indptr_idx = 1, new_data_idx = 0, curr_shape=indptr.shape[0] - 1
        int64_t[::1] new_indptr = np.zeros(new_size + 1, dtype=np.int64)

    with nogil:
        # For each row in the current matrix:
        for curr_row in range(curr_shape):

            # If the row is to be included in the new matrix:
            if bool_mask[curr_row]:

                # Useful quantity to convert the data array index `i` to the
                # equivalent row index in the `bool` mask:
                row_bound = curr_row - indptr[curr_row] + 1

                # For the new indptr array, copy the value from the previous row:
                new_indptr[new_indptr_idx] = new_indptr[new_indptr_idx - 1]

                # For each entry for this row in the data array
                for i in range(indptr[curr_row], indptr[curr_row + 1]):

                    # If the entry isn't filtered, make sure it's included in the new matrix
                    # And increase the `indptr` by one unit:
                    if bool_mask[row_bound + i]:
                        data[new_data_idx] = data[i]
                        new_data_idx += 1
                        new_indptr[new_indptr_idx] += 1

                new_indptr_idx += 1

    return np.asarray(data)[:new_data_idx], np.asarray(new_indptr)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef extract_block_from_ld_data(int[::1] ld_left_bound,
                                 int_dtype[::1] indptr,
                                 noncomplex_numeric[::1] data,
                                 int block_start,
                                 int block_end,
                                 floating dq_scale):
    """
    Given LD data in the form of a CSR matrix represented by LDLinearOperator arrays, this function extracts 
    a block from the matrix. The block is defined by the start and end indices of the rows to be included.
    This implementation assumes that the non-zero elements are contiguous along the diagonal of each row.
    
    The function returns a symmetric numpy matrix of size 
    (block_end - block_start) x (block_end - block_start) containing the data from this block.
    
    :param ld_left_bound: The leftmost column index for each row in the matrix.
    :param indptr: The index pointer array for the CSR matrix to be filtered.
    :param data: The data array for the CSR matrix to be filtered.
    :param block_start: The start index of the block to be extracted.
    :param block_end: The end index of the block to be extracted.
    :param dq_scale: The scaling factor to apply to dequantize the data.
    
    :return: A symmetric numpy matrix containing the data from the block.
    """

    cdef:
        int row_idx, col_idx, row, row_size, col_offset_start, col_offset_end, col_slice_size, block_size = block_end - block_start
        int64_t data_offset
        float[:, ::1] out = np.zeros((block_size, block_size), dtype=np.float32)

    with nogil:

        for row_idx in range(block_size):

            # Determine the row in the original matrix:
            row = block_start + row_idx

            # Determine the size of the row:
            row_size = indptr[row + 1] - indptr[row]

            col_offset_start = row + 1 - ld_left_bound[row]
            col_offset_end = numeric_min(ld_left_bound[row] + row_size, block_end) - ld_left_bound[row]

            col_slice_size = col_offset_end - col_offset_start

            # Compute the offset in the original data array:
            data_offset = indptr[row] + (row + 1) - ld_left_bound[row] + col_offset_start

            # Set the diagonal entry to 1:
            out[row_idx, row_idx] = 1.

            # Set the off-diagonal entries:
            for j in range(col_slice_size):

                # Determine the column index:
                col_idx = row_idx + 1 + j

                # Populate the output array:
                out[row_idx, col_idx] = dq_scale*data[data_offset + j]
                out[col_idx, row_idx] = out[row_idx, col_idx]

    return np.asarray(out)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef slice_ld_data(int[::1] ld_left_bound,
                    int_dtype[::1] indptr,
                    noncomplex_numeric[::1] data,
                    int row_start,
                    int row_end,
                    int col_start,
                    int col_end):
    """
    A general method to slice LD data as represented in LDLinearOperator format, primarily
    the leftmost column index, the index pointer array, and the data array. The function
    take these input arrays and returns equivalent arrays with the sliced data.
    
    :param ld_left_bound: The leftmost column index for each row in the matrix.
    :param indptr: The index pointer array for the CSR matrix to be filtered.
    :param data: The data array for the CSR matrix to be filtered.
    :param row_start: The start index of the rows to be included.
    :param row_end: The end index of the rows to be included.
    :param col_start: The start index of the columns to be included.
    :param col_end: The end index of the columns to be included.
    
    :return: A tuple with the sliced leftmost column index, 
    the sliced index pointer array, and the sliced data array.
    """

    cdef:
        int row_idx, col_idx, row_size, col_offset_start, col_offset_end, col_slice_size, new_indptr_idx = 1
        int_dtype row_offset
        int64_t new_data_idx = 0
        int32_t[::1] new_ld_left_bound = np.zeros(row_end - row_start, dtype=np.int32)
        int64_t[::1] new_indptr = np.zeros(row_end - row_start + 1, dtype=np.int64)
        noncomplex_numeric[::1] new_data = np.empty_like(data, shape=(indptr[row_end] - indptr[row_start], ))

    with nogil:

        for row_idx in range(row_start, row_end):

            # The row offset in the old data array:
            row_offset = indptr[row_idx]
            # The row size in the old data array:
            row_size = indptr[row_idx + 1] - row_offset

            # Determine the offset for the columns depending on the column slice:
            col_offset_start = numeric_max(ld_left_bound[row_idx], col_start) - ld_left_bound[row_idx]
            col_offset_end = numeric_min(ld_left_bound[row_idx] + row_size, col_end) - ld_left_bound[row_idx]

            col_slice_size = col_offset_end - col_offset_start

            # Update the leftmost column index for the row:
            new_ld_left_bound[new_indptr_idx - 1] = col_offset_start + ld_left_bound[row_idx] - row_start
            # Copy the value from the previous row and add the number of columns in the slice:
            new_indptr[new_indptr_idx] = new_indptr[new_indptr_idx - 1] + col_slice_size

            for col_idx in range(col_offset_start, col_offset_end):
                new_data[new_data_idx] = data[row_offset + col_idx]
                new_data_idx += 1

            new_indptr_idx += 1

    return np.asarray(new_data)[:new_data_idx], np.asarray(new_indptr), np.asarray(new_ld_left_bound)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cpdef expand_ranges(int_dtype[::1] start, int_dtype[::1] end, int64_t output_size):
    """
    Given a set of start and end indices, expand them into one long vector that contains 
    the indices between all start and end positions.
    
    :param start: A vector with the start indices.
    :param end: A vector with the end indices.
    :param output_size: The size of the output vector (equivalent to the sum of the lengths
                        of all ranges).
    """

    cdef:
        int_dtype i, j, size=start.shape[0]
        int64_t out_idx = 0
        int_dtype[::1] output = np.empty_like(start, shape=(output_size, ))

    with nogil:
        for i in range(size):
            for j in range(start[i], end[i]):
                output[out_idx] = j
                out_idx += 1

    return np.asarray(output)


# -------------------------------------------------------------------------
# LD Block Boundaries (for LD matrix computation)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef find_ld_block_boundaries(int_dtype[:] pos, int[:, :] block_boundaries):
    """
    Find the LD boundaries for the blockwise estimator of LD, i.e., the 
    indices of the leftmost and rightmost neighbors for each SNP.
    
    :param pos: A vector with the position of each genetic variant.
    :param block_boundaries: A matrix with the boundaries of each LD block.
    """

    cdef:
        int i, j, ldb_idx, B = block_boundaries.shape[0], M = pos.shape[0]
        int_dtype block_start, block_end
        int[:] v_min = np.zeros_like(pos, dtype=np.int32)
        int[:] v_max = M*np.ones_like(pos, dtype=np.int32)

    with nogil:
        for i in range(M):

            # Find the positional boundaries for SNP i:
            for ldb_idx in range(B):
                if block_boundaries[ldb_idx, 0] <= pos[i] < block_boundaries[ldb_idx, 1]:
                    block_start, block_end = block_boundaries[ldb_idx, 0], block_boundaries[ldb_idx, 1]
                    break

            for j in range(i, M):
                if pos[j] >= block_end:
                    v_max[i] = j
                    break

            for j in range(i, -1, -1):
                if pos[j] < block_start:
                    v_min[i] = j + 1
                    break

    return np.array((v_min, v_max))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef find_windowed_ld_boundaries(floating[::1] pos, double max_dist):
    """
    Find the LD boundaries for the windowed estimator of LD, i.e., the 
    indices of the leftmost and rightmost neighbors for each SNP.
    
    .. note::
        To match plink's behavior, the bounds here are inclusive, i.e., 
        if the distance between two SNPs is exactly equal to the maximum distance, 
        they are considered neighbors.
    
    :param pos: A vector with the position of each genetic variant.
    :param max_dist: The maximum distance between SNPs to consider them neighbors.
    """

    cdef:
        int i, j, M = pos.shape[0]
        int[::1] v_min = np.zeros_like(pos, dtype=np.int32)
        int[::1] v_max = M*np.ones_like(pos, dtype=np.int32)

    with nogil:
        for i in range(M):

            for j in range(i, M):
                if pos[j] - pos[i] > max_dist:
                    v_max[i] = j
                    break

            for j in range(i, -1, -1):
                if pos[i] - pos[j] > max_dist:
                    v_min[i] = j + 1
                    break

    return np.array((v_min, v_max))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef find_shrinkage_ld_boundaries(floating[::1] cm_pos,
                                  double genmap_ne,
                                  int genmap_sample_size,
                                  double cutoff):
    """
    Find the LD boundaries for the shrinkage estimator of Wen and Stephens (2010).
    
    :param cm_pos: A vector with the position of each genetic variant in centi Morgan.
    :param genmap_ne: The effective population size for the genetic map sample.
    :param genmap_sample_size: The sample size used to estimate the genetic map.
    :param cutoff: The threshold below which we set the shrinkage factor to zero.
    """

    cdef:
        int i, j, M = cm_pos.shape[0]
        int[::1] v_min = np.zeros_like(cm_pos, dtype=np.int32)
        int[::1] v_max = M*np.ones_like(cm_pos, dtype=np.int32)

    # The multiplicative term for the shrinkage factor
    # The shrinkage factor is 4 * Ne * (rho_ij/100) / (2*m)
    # where Ne is the effective population size and m is the sample size
    # for the genetic map and rho_ij is the distance between SNPs i and j
    # in centi Morgan.
    # Therefore, the multiplicative term that we need to apply
    # to the distance between SNPs is: 4*Ne/(200*m), which is equivalent to 0.02*Ne/m
    # See also: https://github.com/stephenslab/rss/blob/master/misc/get_corr.R
    # and Wen and Stephens (2010)
    cdef double mult_term = 0.02 * genmap_ne / genmap_sample_size

    with nogil:
        for i in range(M):

            for j in range(i, M):
                if exp(-mult_term*(cm_pos[j] - cm_pos[i])) < cutoff:
                    v_max[i] = j
                    break

            for j in range(i, -1, -1):
                if exp(-mult_term*(cm_pos[i] - cm_pos[j])) < cutoff:
                    v_min[i] = j + 1
                    break

    return np.array((v_min, v_max))



import numpy as np
import pandas as pd


def generate_slice_dictionary(vec):
    """
    This utility function takes a sorted vector (e.g. numpy array),
    identifies the unique elements and generates a dictionary of slices
    delineating the start and end positions of each element in the vector.

    :param vec: A numpy array
    :return: A dictionary of slices
    """

    vals, idx = np.unique(vec, return_index=True)
    idx_sort = np.argsort(idx)

    vals = vals[idx_sort]
    idx = idx[idx_sort]

    d = {}

    for i in range(len(idx)):
        try:
            d[vals[i]] = slice(idx[i], idx[i + 1])
        except IndexError:
            d[vals[i]] = slice(idx[i], len(vec))

    return d


def split_block_boundaries(block_iter, max_block_size, mask=None):
    """
    A utility function that takes a list of block boundaries and splits them
    into smaller blocks of a fixed size to satisfy the constraint that no block
    should exceed a certain size.

    :param block_iter: A list of block boundaries (start from 0 and end at the last element).
    :param max_block_size: The maximum size of a block.
    :param mask: A boolean mask to determine which elements are active and will be used in
    the block processing.

    :return: A list of block boundaries that satisfy the constraint.
    """

    new_block_iter = [0]

    for bidx in range(len(block_iter) - 1):

        # Get the block boundaries:
        start = block_iter[bidx]
        end = block_iter[bidx + 1]

        # Determine the current block size:
        if mask is not None:
            original_block_size = mask[start:end].sum()
        else:
            original_block_size = end - start

        if original_block_size > max_block_size:
            # Reverting back to define the blocks in terms of the original indices:
            original_block_size = end - start
            # Determine the number of blocks:
            n_blocks = int(np.ceil(original_block_size / max_block_size))

            new_block_iter += list(new_block_iter[-1] + np.linspace(0,
                                                                    original_block_size,
                                                                    n_blocks + 1, dtype=int)[1:])

        else:
            new_block_iter.append((end - start) + new_block_iter[-1])

    return new_block_iter


def intersect_arrays(arr1, arr2, return_index=False):
    """
    This utility function takes two arrays and returns the shared
    elements (intersection) between them. If return_index is set to True,
    it returns the index of shared elements in the first array.

    :param arr1: The first array
    :param arr2: The second array
    :param return_index: Return the index of shared elements in the first array

    :return: A numpy array of shared elements or their indices
    """

    # NOTE: For best and consistent results, we cast all data types to `str`
    # for now. May need a smarter solution in the future.
    common_elements = pd.DataFrame({'ID': arr1}, dtype=str).reset_index().merge(
        pd.DataFrame({'ID': arr2}, dtype=str)
    )

    if return_index:
        return common_elements['index'].values
    else:
        return common_elements['ID'].values


def intersect_multiple_arrays(arrays):
    """
    This utility function takes multiple arrays and returns the shared
    elements (intersection) between them. This is done with the help of
    the `pd.Index.intersection` method.

    :param arrays: A list of numpy arrays (NOTE: Cannot contain missing values!).
    :return: A numpy array of shared elements.
    """
    common = pd.Index(arrays[0])
    for arr in arrays[1:]:
        common = common.intersection(arr)

    return common.values


def generate_overlapping_windows(seq, window_size, step_size, min_window_size=1):
    """
    Generate overlapping windows of a fixed size over a sequence.

    :param seq: A numpy array of sorted values
    :param window_size: The size of each window.
    :param step_size: The step size between each window. If step_size < window_size, windows will overlap.
    :param min_window_size: The minimum size of a window. Windows smaller than this size will be discarded.

    :return: A numpy array of start and end indices of each window.
    """

    # Calculate the start of each window
    starts = np.arange(seq[0], seq[-1] - window_size, step_size)

    # Find the indices where each window starts and ends
    start_indices = np.searchsorted(seq, starts)
    end_indices = np.searchsorted(seq, starts + window_size, side='right')

    # Ensure that the last window is not truncated and split it if necessary:
    if end_indices[-1] < seq.shape[0]:
        end_indices[-1] = seq.shape[0]

    # Combine start and end indices:
    block_iter = np.column_stack((start_indices, end_indices))
    # Filter to keep only valid blocks:
    block_iter = block_iter[block_iter[:, 0] + min_window_size <= block_iter[:, 1], :]

    return block_iter


def detect_header_keywords(fname, keywords):
    """
    Detect if the first line of a file contains any of the provided keywords.
    This is used to check for whether headers are present in a file.

    :param fname: The fpath to the file
    :param keywords: A string or list of strings representing keywords to search for.

    :return: True if any of the keywords are found, False otherwise.
    """

    if isinstance(keywords, str):
        keywords = [keywords]

    with open(fname, 'r') as f:
        line = f.readline().strip()

    return any([kw in line for kw in keywords])


def is_numeric(obj):
    """
    Check if a python object is numeric. This function handles
    numpy arrays and scalars.
    :param obj: A python object
    :return: True if the object is numeric, False otherwise.
    """
    if isinstance(obj, np.ndarray):
        return np.issubdtype(obj.dtype, np.number)
    else:
        return np.issubdtype(type(obj), np.number)


def iterable(arg):
    """
    Check if an object is iterable, but not a string.
    :param arg: A python object.
    :return: True if the object is iterable, False otherwise.
    """

    import collections.abc

    return (
        isinstance(arg, collections.abc.Iterable)
        and not isinstance(arg, str)
    )
